from pathlib import Path
from typing import Union

import numpy as np


def export_points_to_brainrender(
    points: np.ndarray,
    resolution: float,
    output_filename: Union[str, Path],
) -> None:
    """
    Export points in atlas space for visualization in brainrender.

    Points are scaled from atlas coordinates to real units and saved
    as a numpy file.

    Parameters
    ----------
    points : np.ndarray
        A numpy array containing the points in atlas space.
    resolution : float
        A numerical value representing the resolution scale to be
        applied to the points.
    output_filename : Union[str, Path]
        The path where the numpy file will be saved. Can be a string
        or a Path object.

    Returns
    -------
    None
    """
    np.save(output_filename, points * resolution)



import json
from pathlib import Path
from typing import Any, Dict, List, Union

import napari
import pandas as pd
import tifffile
from brainglobe_atlasapi import BrainGlobeAtlas
from brainglobe_atlasapi.list_atlases import get_downloaded_atlases
from brainglobe_space import AnatomicalSpace
from qtpy import QtCore
from qtpy.QtWidgets import (
    QComboBox,
    QFileDialog,
    QGridLayout,
    QLabel,
    QTableView,
    QWidget,
)

from brainglobe_utils.brainmapper.analysis import (
    summarise_points_by_atlas_region,
)
from brainglobe_utils.brainmapper.export import export_points_to_brainrender
from brainglobe_utils.brainreg.transform import (
    transform_points_from_downsampled_to_atlas_space,
)
from brainglobe_utils.general.system import ensure_extension
from brainglobe_utils.qtpy.dialog import display_info
from brainglobe_utils.qtpy.interaction import add_button, add_combobox
from brainglobe_utils.qtpy.logo import header_widget
from brainglobe_utils.qtpy.table import DataFrameModel


class TransformPoints(QWidget):
    def __init__(self, viewer: napari.viewer.Viewer):
        """
        Initialize the TransformPoints widget.

        Parameters
        ----------
        viewer : napari.viewer.Viewer
            The napari viewer instance.
            This will be passed when opening the widget.
        """
        super(TransformPoints, self).__init__()
        self.viewer = viewer
        self.raw_data = None
        self.points_layer = None
        self.atlas = None
        self.transformed_points = None

        self.image_layer_names = self._get_layer_names(
            layer_type=napari.layers.Image
        )
        self.points_layer_names = self._get_layer_names(
            layer_type=napari.layers.Points
        )
        self.setup_main_layout()

        @self.viewer.layers.events.connect
        def update_layer_list(v: napari.viewer.Viewer) -> None:
            """
            Update internal list of layers whenever the napari layers list
            is updated.

            Parameters
            ----------
            v : napari.viewer.Viewer
                The napari viewer instance.
            """
            self.image_layer_names = self._get_layer_names(
                layer_type=napari.layers.Image
            )
            self.points_layer_names = self._get_layer_names(
                layer_type=napari.layers.Points
            )

            self._update_combobox_options(
                self.raw_data_choice, self.image_layer_names
            )

            self._update_combobox_options(
                self.points_layer_choice, self.points_layer_names
            )

    @staticmethod
    def _update_combobox_options(
        combobox: QComboBox, options_list: List[str]
    ) -> None:
        """
        Update the options in a QComboBox.

        Parameters
        ----------
        combobox : QComboBox
            The combobox to update.
        options_list : List[str]
            The list of options to set in the combobox.
        """

        original_text = combobox.currentText()
        combobox.clear()
        combobox.addItems(options_list)
        combobox.setCurrentText(original_text)

    def _get_layer_names(
        self,
        layer_type: napari.layers.Layer,
        default: str = "",
    ) -> List[str]:
        """
        Get list of layer names of a given layer type.

        Parameters
        ----------
        layer_type : napari.layers.Layer, optional
            The type of layer to get names for.
        default : str, optional
            Default values to include in the list. Default is an empty string.

        Returns
        -------
        List[str]
            A list of layer names.
        """
        layer_names = [
            layer.name
            for layer in self.viewer.layers
            if isinstance(layer, layer_type)
        ]

        if layer_names:
            return [default] + layer_names
        else:
            return [default]

    def setup_main_layout(self) -> None:
        """
        Set up the main layout of the widget.
        """
        self.layout = QGridLayout()
        self.layout.setContentsMargins(10, 10, 10, 10)
        self.layout.setAlignment(QtCore.Qt.AlignTop)
        self.layout.setSpacing(4)
        self.add_header()
        self.add_points_combobox(row=1, column=0)
        self.add_raw_data_combobox(row=2, column=0)
        self.add_transform_button(row=3, column=0)
        self.add_brainrender_export_button(row=3, column=1)
        self.add_points_summary_table(row=4, column=0)
        self.add_save_all_points_button(row=6, column=0)
        self.add_save_points_summary_button(row=6, column=1)
        self.add_status_label(row=7, column=0)

        self.setLayout(self.layout)

    def add_header(self) -> None:
        """
        Header including brainglobe logo and documentation links.
        """
        header = header_widget(
            package_name="brainmapper",
            package_tagline="Transform points to atlas space",
            github_repo_name="brainglobe-utils",
            citation_doi="https://doi.org/10.1371/journal.pcbi.1009074",
        )
        self.layout.addWidget(header, 0, 0, 1, 2)

    def add_points_combobox(self, row: int, column: int) -> None:
        """
        Add a combobox for selecting the points layer containing
        the points (e.g. cells) to transform to a BrainGlobe atlas.

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.points_layer_choice, _ = add_combobox(
            self.layout,
            "Points layer",
            self.points_layer_names,
            column=column,
            row=row,
            callback=self.set_points_layer,
        )

    def add_raw_data_combobox(self, row: int, column: int) -> None:
        """
        Add a combobox for selecting the raw data layer. This defines
        the coordinate space for transforming points to a BrainGlobe atlas.

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.raw_data_choice, _ = add_combobox(
            self.layout,
            "Raw data layer",
            self.image_layer_names,
            column=column,
            row=row,
            callback=self.set_raw_data_layer,
        )

    def add_transform_button(self, row: int, column: int) -> None:
        """
        Add a button to begin the transformation of points to atlas space.

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.transform_button = add_button(
            "Transform points",
            self.layout,
            self.transform_points_to_atlas_space,
            row=row,
            column=column,
            visibility=True,
            tooltip="Transform points layer to atlas space",
        )

    def add_brainrender_export_button(self, row: int, column: int) -> None:
        """
        Add a button to export the points in atlas space in the brainrender
        format.

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.brainrender_export_button = add_button(
            "Export to brainrender",
            self.layout,
            self.export_points_to_brainrender,
            row=row,
            column=column,
            visibility=False,
            tooltip="Export points in atlas space to brainrender",
        )

    def add_points_summary_table(self, row: int, column: int) -> None:
        """
        Add a table to display the summary of points per atlas region.

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.points_per_region_table_title = QLabel(
            "Points distribution summary"
        )
        self.points_per_region_table_title.setVisible(False)
        self.layout.addWidget(self.points_per_region_table_title, row, column)
        self.points_per_region_table = QTableView()
        self.points_per_region_table.setVisible(False)
        self.layout.addWidget(self.points_per_region_table, row + 1, column)

    def add_save_all_points_button(self, row: int, column: int) -> None:
        """
        Add a button to save all points information (i.e. the list of
        all points, and their coordinates in raw data and atlas
        space, alongside assigned atlas region).

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.save_all_points_button = add_button(
            "Save all points information",
            self.layout,
            self.save_all_points_csv,
            row=row,
            column=column,
            visibility=False,
            tooltip="Save all points information as a csv file",
        )

    def add_save_points_summary_button(self, row: int, column: int) -> None:
        """
        Add a button to save points summary (i.e. points per atlas region).

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.save_points_summary_button = add_button(
            "Save points summary",
            self.layout,
            self.save_points_summary_csv,
            row=row,
            column=column,
            visibility=False,
            tooltip="Save points summary as a csv file",
        )

    def add_status_label(self, row: int, column: int) -> None:
        """
        Add a status label to inform the user of progress.

        Parameters
        ----------
        row : int
            Row in the grid layout.
        column : int
            Column in the grid layout.
        """
        self.status_label = QLabel()
        self.status_label.setText("Ready")
        self.layout.addWidget(self.status_label, row, column)

    def set_raw_data_layer(self) -> None:
        """
        Set background layer from current background text box selection.
        """
        if self.raw_data_choice.currentText() != "":
            self.raw_data = self.viewer.layers[
                self.raw_data_choice.currentText()
            ]

    def set_points_layer(self) -> None:
        """
        Set background layer from current background text box selection.
        """
        if self.points_layer_choice.currentText() != "":
            self.points_layer = self.viewer.layers[
                self.points_layer_choice.currentText()
            ]

    def transform_points_to_atlas_space(self) -> None:
        """
        Transform points layer to atlas space.
        """
        layers_in_place = self.check_layers()
        if not layers_in_place:
            return

        self.status_label.setText("Loading brainreg data ...")
        data_loaded = self.load_brainreg_directory()

        if not data_loaded:
            self.status_label.setText("Ready")
            return

        self.status_label.setText("Transforming points ...")

        self.run_transform_points_to_downsampled_space()
        self.run_transform_downsampled_points_to_atlas_space()

        self.status_label.setText("Analysing point distribution ...")
        self.analyse_points()
        self.status_label.setText("Ready")

    def check_layers(self) -> bool:
        """
        Check if the layers needed to begin the transformation
        have been selected by the user.

        Returns
        -------
        bool
            True if both raw data and points layers are selected,
            False otherwise.
        """
        if self.raw_data is None and self.points_layer is None:
            display_info(
                self,
                "No layers selected",
                "Please select the layers corresponding to the points "
                "you would like to transform and the raw data (registered by "
                "brainreg)",
            )
            return False

        if self.raw_data is None:
            display_info(
                self,
                "No raw data layer selected",
                "Please select a layer that corresponds to the raw "
                "data (registered by brainreg)",
            )
            return False

        if self.points_layer is None:
            display_info(
                self,
                "No points layer selected",
                "Please select a points layer you would like to transform",
            )
            return False

        return True

    def load_brainreg_directory(self) -> bool:
        """
        Load the brainreg directory selected by the user.
        Returns false if not selected, to abort analysis.

        Returns
        -------
        bool
            True if a directory was selected, False otherwise.
        """
        brainreg_directory = QFileDialog.getExistingDirectory(
            self,
            "Select brainreg directory",
        )
        if brainreg_directory == "":
            return False
        else:
            self.brainreg_directory = Path(brainreg_directory)

        self.initialise_brainreg_data()
        self.status_label.setText("Ready")
        return True

    def initialise_brainreg_data(self) -> None:
        """
        Initialize brainreg data by defining the paths,
        then loading the brainreg metadata, and then the atlas.
        """
        self.paths = Paths(self.brainreg_directory)
        self.check_brainreg_directory()
        self.metadata = Metadata(self.brainreg_metadata)
        self.load_atlas()

    def check_brainreg_directory(self) -> None:
        """
        Check if the selected directory is a valid brainreg directory
        by checking for the existence of an "atlas" entry in the json
        """
        try:
            with open(self.paths.brainreg_metadata_file) as json_file:
                self.brainreg_metadata = json.load(json_file)

                if "atlas" not in self.brainreg_metadata:
                    self.display_brainreg_directory_warning()

        except FileNotFoundError:
            self.display_brainreg_directory_warning()

    def display_brainreg_directory_warning(self) -> None:
        """
        Display a warning to the user if the selected directory
        is not a valid brainreg directory.
        """
        display_info(
            self,
            "Not a brainreg directory",
            "This directory does not appear to be a valid brainreg "
            "directory. Please try loading another brainreg output directory.",
        )

    def load_atlas(self) -> None:
        """
        Load the BrainGlobe atlas used for the initial brainreg registration.
        """
        if not self.is_atlas_installed(self.metadata.atlas_string):
            display_info(
                self,
                "Atlas not downloaded",
                f"Atlas: {self.metadata.atlas_string} needs to be "
                f"downloaded. This may take some time depending on "
                f"the size of the atlas and your network speed.",
            )
        self.atlas = BrainGlobeAtlas(self.metadata.atlas_string)

    def run_transform_points_to_downsampled_space(self) -> None:
        """
        Transform points fromm the raw data space (in which points
        were detected) to the downsampled space defined by brainreg.
        This space is the same as the raw data, but downsampled and realigned
        to match the orientation and resolution of the atlas.
        """
        downsampled_space = self.get_downsampled_space()
        raw_data_space = self.get_raw_data_space()
        self.points_in_downsampled_space = raw_data_space.map_points_to(
            downsampled_space, self.points_layer.data
        )
        self.viewer.add_points(
            self.points_in_downsampled_space,
            name="Points in downsampled space",
            visible=False,
        )

    def run_transform_downsampled_points_to_atlas_space(self) -> None:
        """
        Transform points from the downsampled space to atlas space. Uses
        the deformation fields output by NiftyReg (via brainreg) as a look up.
        """
        deformation_field_paths = [
            self.paths.deformation_field_0,
            self.paths.deformation_field_1,
            self.paths.deformation_field_2,
        ]
        self.points_in_atlas_space, points_out_of_bounds = (
            transform_points_from_downsampled_to_atlas_space(
                self.points_in_downsampled_space,
                self.atlas,
                deformation_field_paths,
                warn_out_of_bounds=False,
            )
        )
        self.viewer.add_points(
            self.points_in_atlas_space,
            name="Points in atlas space",
            visible=True,
        )

        if len(points_out_of_bounds) > 0:
            display_info(
                self,
                "Points outside atlas",
                f"{len(points_out_of_bounds)} "
                f"points fell outside the atlas space",
            )

    def get_downsampled_space(self) -> AnatomicalSpace:
        """
        Get the anatomical space (as defined by brainglobe-space)
        for the downsampled data.

        Returns
        -------
        AnatomicalSpace
            The downsampled anatomical space (as defined by brainglobe-space).
        """
        target_shape = tifffile.imread(self.paths.downsampled_image).shape

        downsampled_space = AnatomicalSpace(
            self.atlas.orientation,
            shape=target_shape,
            resolution=self.atlas.resolution,
        )
        return downsampled_space

    def get_raw_data_space(self) -> AnatomicalSpace:
        """
        Get the anatomical space (as defined by brainglobe-space)
        for the raw data.

        Returns
        -------
        AnatomicalSpace
            The raw data anatomical space (as defined by brainglobe-space).
        """
        raw_data_space = AnatomicalSpace(
            self.metadata.orientation,
            shape=self.raw_data.data.shape,
            resolution=[float(i) for i in self.metadata.voxel_sizes],
        )
        return raw_data_space

    def analyse_points(self) -> None:
        """
        Analyse the distribution of points in the space
        of the BrainGlobe Atlas.
        """
        self.all_points_df, self.points_per_region_df = (
            summarise_points_by_atlas_region(
                self.points_layer.data,
                self.points_in_atlas_space,
                self.atlas,
                self.paths.volume_csv_path,
            )
        )

        self.populate_summary_table()
        self.brainrender_export_button.setVisible(True)
        self.save_all_points_button.setVisible(True)
        self.save_points_summary_button.setVisible(True)

        print("Analysing points")

    def populate_summary_table(
        self,
        columns_to_keep: List[str] = [
            "structure_name",
            "left_cell_count",
            "right_cell_count",
        ],
    ) -> None:
        """
        Populate the table with the summary of points per atlas region.

        Parameters
        ----------
        columns_to_keep : List[str], optional
            Columns to keep in the summary table.
            Default columns are ["structure_name", "left_cell_count",
            "right_cell_count"].
        """
        summary_df = self.points_per_region_df[columns_to_keep]
        self.points_per_region_table_model = DataFrameModel(summary_df)
        self.points_per_region_table.setModel(
            self.points_per_region_table_model
        )
        self.points_per_region_table_title.setVisible(True)
        self.points_per_region_table.setVisible(True)

    def export_points_to_brainrender(self) -> None:
        """
        Export points in the format required for brainrender
        N.B. assumes atlas is isotropic
        """
        path, _ = QFileDialog.getSaveFileName(
            self,
            "Choose filename",
            "",
            "NumPy Files (*.npy)",
        )

        if path:
            path = ensure_extension(path, ".npy")
        export_points_to_brainrender(
            self.points_in_atlas_space, self.atlas.resolution[0], path
        )

    def save_all_points_csv(self) -> None:
        """
        Save the coordinate and atlas region
        information for all points to a CSV file.
        """
        self.save_df_to_csv(self.all_points_df)

    def save_points_summary_csv(self) -> None:
        """
        Save the summary of the distribution of points
        in the atlas space to a CSV file.
        """
        self.save_df_to_csv(self.points_per_region_df)

    def save_df_to_csv(self, df: pd.DataFrame) -> None:
        """
        Save the given DataFrame to a CSV file.

        Prompts the user to choose a filename and ensures the file has a
        .csv extension.
        The DataFrame is then saved to the specified file.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to be saved.

        Returns
        -------
        None
        """
        path, _ = QFileDialog.getSaveFileName(
            self,
            "Choose filename",
            "",
            "CSV Files (*.csv)",
        )

        if path:
            path = ensure_extension(path, ".csv")
            df.to_csv(path, index=False)

    @staticmethod
    def is_atlas_installed(atlas: str) -> bool:
        """
        Check if the specified BrainGlobe atlas is installed.

        Parameters
        ----------
        atlas : str
            The atlas name to check.

        Returns
        -------
        bool
            True if the atlas is installed, False otherwise.
        """
        downloaded_atlases = get_downloaded_atlases()
        if atlas in downloaded_atlases:
            return True
        else:
            return False


class Paths:
    """
    A class to hold all brainreg-related file paths.

    N.B. this could be imported from brainreg, but it is copied here to
    prevent a circular dependency

    Attributes
    ----------
    brainreg_directory : Path
        Path to brainreg output directory (or brainmapper
        "registration" directory)
    brainreg_metadata_file : Path
        The path to the brainreg metadata (brainreg.json) file
    deformation_field_0 : Path
        The path to the deformation field (0th dimension)
    deformation_field_1 : Path
        The path to the deformation field (1st dimension)
    deformation_field_2 : Path
        The path to the deformation field (2nd dimension)
    downsampled_image : Path
        The path to the downsampled.tiff image file
    volume_csv_path : Path
        The path to the csv file containing region volumes

    Parameters
    ----------
    brainreg_directory : Union[str, Path]
        Path to brainreg output directory (or brainmapper
        "registration" directory)
    """

    def __init__(self, brainreg_directory: Union[str, Path]) -> None:
        """
        Set the paths based on the given brainreg directory

        Parameters
        ----------
        brainreg_directory : Union[str, Path]
            Path to brainreg output directory
            (or brainmapper "registration" directory).
        """
        self.brainreg_directory: Path = Path(brainreg_directory)
        self.brainreg_metadata_file: Path = self.make_filepaths(
            "brainreg.json"
        )
        self.deformation_field_0: Path = self.make_filepaths(
            "deformation_field_0.tiff"
        )
        self.deformation_field_1: Path = self.make_filepaths(
            "deformation_field_1.tiff"
        )
        self.deformation_field_2: Path = self.make_filepaths(
            "deformation_field_2.tiff"
        )
        self.downsampled_image: Path = self.make_filepaths("downsampled.tiff")
        self.volume_csv_path: Path = self.make_filepaths("volumes.csv")

    def make_filepaths(self, filename: str) -> Path:
        """
        Create a full file path by combining the directory with a filename.

        Parameters
        ----------
        filename : str
            The name of the file to create a path for.

        Returns
        -------
        Path
            The full path to the specified file.
        """
        return self.brainreg_directory / filename


class Metadata:
    """
    A class to represent brainreg registration metadata
    (loaded from brainreg.json)

    Attributes
    ----------
    orientation : str
        The orientation of the input data (in brainglobe-space format)
    atlas_string : str
        The BrainGlobe atlas used for brain registration.
    voxel_sizes : List[float]
        The voxel sizes of the input data

    Parameters
    ----------
    brainreg_metadata : Dict[str, Any]
        A dictionary containing metadata information,
        loaded from brainreg.json
    """

    def __init__(self, brainreg_metadata: Dict[str, Any]) -> None:
        """
        Initialize the Metadata instance with brainreg metadata.

        Parameters
        ----------
        brainreg_metadata : Dict[str, Any]
            A dictionary containing metadata information from brainreg.json
        """
        self.orientation: str = brainreg_metadata["orientation"]
        self.atlas_string: str = brainreg_metadata["atlas"]
        self.voxel_sizes: List[float] = brainreg_metadata["voxel_sizes"]



import glob
import logging
import math
import os
import warnings
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Tuple

import numpy as np
import tifffile
from dask import array as da
from dask import delayed
from natsort import natsorted
from skimage import transform
from tqdm import tqdm

from brainglobe_utils.general.system import (
    get_num_processes,
    get_sorted_file_paths,
)
from brainglobe_utils.IO.image.utils import ImageIOLoadException

from .utils import check_mem, scale_z

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import nibabel as nib


def load_any(
    src_path,
    x_scaling_factor=1.0,
    y_scaling_factor=1.0,
    z_scaling_factor=1.0,
    anti_aliasing=True,
    load_parallel=False,
    sort_input_file=False,
    as_numpy=False,
    n_free_cpus=2,
):
    """
    This function will guess the type of data and hence call the appropriate
    function from this module to load the given brain.

    .. warning:: x and y scaling not used at the moment if loading a
        complete image

    Parameters
    ----------
    src_path : str or pathlib.Path
        Can be the path of a nifty file, tiff file, tiff files
        folder, or text file containing a list of paths.

    x_scaling_factor : float, optional
        The scaling of the brain along the x dimension (applied on loading
        before return).

    y_scaling_factor : float, optional
        The scaling of the brain along the y dimension (applied on loading
        before return).

    z_scaling_factor : float, optional
        The scaling of the brain along the z dimension (applied on loading
        before return).

    anti_aliasing : bool, optional
        Whether to apply a Gaussian filter to smooth the image prior to
        down-scaling. It is crucial to filter when down-sampling the image to
        avoid aliasing artifacts.

    load_parallel : bool, optional
        Load planes in parallel using multiprocessing for faster data loading.

    sort_input_file : bool, optional
        If set to true and the input is a filepaths file, it will be naturally
        sorted.

    as_numpy : bool, optional
        Whether to convert the image to a numpy array in memory (rather than a
        memmap). Only relevant for .nii files.

    verbose : bool, optional
        Print more information about the process.

    n_free_cpus : int, optional
        Number of CPU cores to leave free.

    Returns
    -------
    np.ndarray
        The loaded brain.

    Raises
    ------
    ImageIOLoadException
        If there was an issue loading the image with image.

    See Also
    ------
    image.utils.ImageIOLoadException
    """
    src_path = Path(src_path)

    if src_path.is_dir():
        logging.debug("Data type is: directory of files")
        img = load_from_folder(
            src_path,
            x_scaling_factor,
            y_scaling_factor,
            z_scaling_factor,
            anti_aliasing=anti_aliasing,
            file_extension=".tif*",
            load_parallel=load_parallel,
            n_free_cpus=n_free_cpus,
        )
    elif src_path.suffix == ".txt":
        logging.debug("Data type is: list of file paths")
        img = load_img_sequence(
            src_path,
            x_scaling_factor,
            y_scaling_factor,
            z_scaling_factor,
            anti_aliasing=anti_aliasing,
            load_parallel=load_parallel,
            sort=sort_input_file,
            n_free_cpus=n_free_cpus,
        )
    elif src_path.suffix in [".tif", ".tiff"]:
        logging.debug("Data type is: tif stack")
        img = load_img_stack(
            src_path,
            x_scaling_factor,
            y_scaling_factor,
            z_scaling_factor,
            anti_aliasing=anti_aliasing,
        )
    elif src_path.suffix == ".nii" or str(src_path).endswith(".nii.gz"):
        logging.debug("Data type is: NifTI")
        img = load_nii(src_path, as_array=True, as_numpy=as_numpy)
    else:
        raise NotImplementedError(
            "Could not guess data type for path {}".format(src_path)
        )

    return img


def load_img_stack(
    stack_path,
    x_scaling_factor,
    y_scaling_factor,
    z_scaling_factor,
    anti_aliasing=True,
):
    """
    Load a tiff stack as a numpy array.

    Parameters
    ----------
    stack_path : str or pathlib.Path
        The path of the image to be loaded. Note that only 3D tiffs are
        supported.

    x_scaling_factor : float
        The scaling of the brain along the x dimension (applied on loading
        before return).

    y_scaling_factor : float
        The scaling of the brain along the y dimension (applied on loading
        before return).

    z_scaling_factor : float
        The scaling of the brain along the z dimension (applied on loading
        before return).

    anti_aliasing : bool, optional
        Whether to apply a Gaussian filter to smooth the image prior to
        down-scaling. It is crucial to filter when down-sampling the image to
        avoid aliasing artifacts.

    Returns
    -------
    np.ndarray
        The loaded brain array.

    Raises
    ------
    ImageIOLoadException
        If attempt to load a 2D tiff.
    """
    stack_path = Path(stack_path)
    logging.debug(f"Loading: {stack_path}")
    stack = tifffile.imread(stack_path)

    if stack.ndim != 3:
        raise ImageIOLoadException(error_type="2D tiff")

    # Downsampled plane by plane because the 3D downsampling in scipy etc
    # uses too much RAM

    if not (x_scaling_factor == y_scaling_factor == 1):
        downsampled_stack = []
        logging.debug("Downsampling stack in X/Y")
        for plane in tqdm(range(0, len(stack))):
            downsampled_stack.append(
                transform.rescale(
                    stack[plane],
                    (y_scaling_factor, x_scaling_factor),
                    mode="constant",
                    preserve_range=True,
                    anti_aliasing=anti_aliasing,
                )
            )

        logging.debug("Converting downsampled stack to array")
        stack = np.array(downsampled_stack)

    if z_scaling_factor != 1:
        logging.debug("Downsampling stack in Z")
        stack = scale_z(stack, z_scaling_factor)
    return stack


def load_nii(src_path, as_array=False, as_numpy=False):
    """
    Load a brain from a nifti file.

    Parameters
    ----------
    src_path : str or pathlib.Path
        The path to the nifty file on the filesystem.

    as_array : bool, optional
        Whether to convert the brain to a numpy array or keep it as a nifty
        object.

    as_numpy : bool, optional
        Whether to convert the image to a numpy array in memory (rather than a
        memmap).

    Returns
    -------
    np.ndarray or nifty object
        The loaded brain. The format depends on the `as_array` flag.
    """
    src_path = Path(src_path)
    nii_img = nib.load(src_path)
    if as_array:
        image = nii_img.get_fdata()
        if as_numpy:
            image = np.array(image)

        return image
    else:
        return nii_img


def load_from_folder(
    src_folder,
    x_scaling_factor=1,
    y_scaling_factor=1,
    z_scaling_factor=1,
    anti_aliasing=True,
    file_extension="",
    load_parallel=False,
    n_free_cpus=2,
):
    """
    Load a brain from a folder. All tiff files will be read sorted and assumed
    to belong to the same sample. Optionally, a name_filter string can be
    supplied which will have to be present in the file names for them to be
    considered part of the sample.

    Parameters
    ----------
    src_folder : str or pathlib.Path
        The source folder containing tiff files. Note that this folder must
        contain at least 2 tiffs, and all tiff images must have the same shape.

    x_scaling_factor : float, optional
        The scaling of the brain along the x dimension (applied on loading
        before return).

    y_scaling_factor : float, optional
        The scaling of the brain along the y dimension (applied on loading
        before return).

    z_scaling_factor : float, optional
        The scaling of the brain along the z dimension.

    anti_aliasing : bool, optional
        Whether to apply a Gaussian filter to smooth the image prior to
        down-scaling. It is crucial to filter when down-sampling the image to
        avoid aliasing artifacts.

    file_extension : str, optional
        Will have to be present in the file names for them to be considered
        part of the sample.

    load_parallel : bool, optional
        Use multiprocessing to speed up image loading.

    n_free_cpus : int, optional
        Number of CPU cores to leave free.

    Returns
    -------
    np.ndarray
        The loaded and scaled brain.

    Raises
    ------
    ImageIOLoadException
        If attempt to load a directory containing only a single tiff, or a
        sequence of tiffs that have different shapes.
    """
    paths = get_sorted_file_paths(src_folder, file_extension=file_extension)

    return load_image_series(
        paths,
        x_scaling_factor,
        y_scaling_factor,
        z_scaling_factor,
        load_parallel=load_parallel,
        n_free_cpus=n_free_cpus,
        anti_aliasing=anti_aliasing,
    )


def load_img_sequence(
    img_sequence_file_path,
    x_scaling_factor=1,
    y_scaling_factor=1,
    z_scaling_factor=1,
    anti_aliasing=True,
    load_parallel=False,
    sort=False,
    n_free_cpus=2,
):
    """
    Load a brain from a sequence of files specified in a text file containing
    an ordered list of paths.

    Parameters
    ----------
    img_sequence_file_path : str or pathlib.Path
        The path to the file containing the ordered list of image paths (one
        per line). Note that this file must contain at least 2 paths, and all
        referenced images must have the same shape.

    x_scaling_factor : float, optional
        The scaling of the brain along the x dimension (applied on loading
        before return).

    y_scaling_factor : float, optional
        The scaling of the brain along the y dimension (applied on loading
        before return).

    z_scaling_factor : float, optional
        The scaling of the brain along the z dimension.

    anti_aliasing : bool, optional
        Whether to apply a Gaussian filter to smooth the image prior to
        down-scaling. It is crucial to filter when down-sampling the image to
        avoid aliasing artifacts.

    load_parallel : bool, optional
        Use multiprocessing to speed up image loading.

    sort : bool, optional
        If set to true, will perform a natural sort of the file paths in the
        list.

    n_free_cpus : int, optional
        Number of CPU cores to leave free.

    Returns
    -------
    np.ndarray
        The loaded and scaled brain.

    Raises
    ------
    ImageIOLoadException
        If attempt to load a txt file containing only a single path, or a
        sequence of paths that load images with different shapes.
    """
    img_sequence_file_path = Path(img_sequence_file_path)
    with open(img_sequence_file_path, "r") as in_file:
        paths = in_file.readlines()
        paths = [p.strip() for p in paths]
    if sort:
        paths = natsorted(paths)

    return load_image_series(
        paths,
        x_scaling_factor,
        y_scaling_factor,
        z_scaling_factor,
        load_parallel=load_parallel,
        n_free_cpus=n_free_cpus,
        anti_aliasing=anti_aliasing,
    )


def load_image_series(
    paths,
    x_scaling_factor=1,
    y_scaling_factor=1,
    z_scaling_factor=1,
    anti_aliasing=True,
    load_parallel=False,
    n_free_cpus=2,
):
    """
    Load a brain from a sequence of image paths.

    Parameters
    ----------
    paths : list of str or list of pathlib.Path
        Ordered list of image paths. Must contain at least 2 paths, and all
        referenced images must have the same shape.

    x_scaling_factor : float, optional
        The scaling of the brain along the x dimension (applied on loading
        before return).

    y_scaling_factor : float, optional
        The scaling of the brain along the y dimension (applied on loading
        before return).

    z_scaling_factor : float, optional
        The scaling of the brain along the z dimension.

    anti_aliasing : bool, optional
        Whether to apply a Gaussian filter to smooth the image prior to
        down-scaling. It is crucial to filter when down-sampling the image to
        avoid aliasing artifacts.

    load_parallel : bool, optional
        Use multiprocessing to speed up image loading.

    n_free_cpus : int, optional
        Number of CPU cores to leave free.

    Returns
    -------
    np.ndarray
        The loaded and scaled brain.

    Raises
    ------
    ImageIOLoadException
        If attempt to load a single path, or a sequence of paths that load
        images with different shapes.
    """
    # Throw an error if there's only one image to load - should be an image
    # series, so at least 2 paths.
    if len(paths) == 1:
        raise ImageIOLoadException("single_tiff")

    if load_parallel:
        img = threaded_load_from_sequence(
            paths,
            x_scaling_factor,
            y_scaling_factor,
            n_free_cpus=n_free_cpus,
            anti_aliasing=anti_aliasing,
        )
    else:
        img = load_from_paths_sequence(
            paths,
            x_scaling_factor,
            y_scaling_factor,
            anti_aliasing=anti_aliasing,
        )

    img = np.moveaxis(img, 2, 0)  # back to z first
    if z_scaling_factor != 1:
        img = scale_z(img, z_scaling_factor)

    return img


def threaded_load_from_sequence(
    paths_sequence,
    x_scaling_factor=1.0,
    y_scaling_factor=1.0,
    anti_aliasing=True,
    n_free_cpus=2,
):
    """
    Use multiprocessing to load a brain from a sequence of image paths.

    Parameters
    ----------
    paths_sequence : list of str or list of pathlib.Path
        The sorted list of the planes paths on the filesystem. All planes
        must have the same shape.

    x_scaling_factor : float, optional
        The scaling of the brain along the x dimension (applied on loading
        before return).

    y_scaling_factor : float, optional
        The scaling of the brain along the y dimension (applied on loading
        before return).

    anti_aliasing : bool, optional
        Whether to apply a Gaussian filter to smooth the image prior to
        down-scaling. It is crucial to filter when down-sampling the image to
        avoid aliasing artifacts.

    n_free_cpus : int, optional
        Number of CPU cores to leave free.

    Returns
    -------
    np.ndarray
        The loaded and scaled brain.

    Raises
    ------
    ImageIOLoadException
        If attempt to load a sequence of images with different shapes.
    """
    stacks = []
    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)

    # WARNING: will not work with interactive interpreter.
    pool = ProcessPoolExecutor(max_workers=n_processes)
    # FIXME: should detect and switch to other method

    n_paths_per_subsequence = math.ceil(len(paths_sequence) / n_processes)
    for i in range(n_processes):
        start_idx = i * n_paths_per_subsequence
        if start_idx >= len(paths_sequence):
            break
        else:
            end_idx = start_idx + n_paths_per_subsequence

            if end_idx <= len(paths_sequence):
                sub_paths = paths_sequence[start_idx:end_idx]
            else:
                sub_paths = paths_sequence[start_idx:]

        process = pool.submit(
            load_from_paths_sequence,
            sub_paths,
            x_scaling_factor,
            y_scaling_factor,
            anti_aliasing=anti_aliasing,
        )
        stacks.append(process)

    stack_shapes = set()
    for i in range(len(stacks)):
        stacks[i] = stacks[i].result()
        stack_shapes.add(stacks[i].shape[0:2])

    # Raise an error if the x/y shape of all stacks aren't the same
    if len(stack_shapes) > 1:
        raise ImageIOLoadException("sequence_shape")

    stack = np.dstack(stacks)
    return stack


def load_from_paths_sequence(
    paths_sequence,
    x_scaling_factor=1.0,
    y_scaling_factor=1.0,
    anti_aliasing=True,
):
    # TODO: Optimise -  load threaded and process by batch
    """
    A single core version of the function to load a brain from a sequence of
    image paths.

    Parameters
    ----------
    paths_sequence : list of str or list of pathlib.Path
        The sorted list of the planes paths on the filesystem. All planes
        must have the same shape.

    x_scaling_factor : float, optional
        The scaling of the brain along the x dimension (applied on loading
        before return).

    y_scaling_factor : float, optional
        The scaling of the brain along the y dimension (applied on loading
        before return).

    anti_aliasing : bool, optional
        Whether to apply a Gaussian filter to smooth the image prior to
        down-scaling. It is crucial to filter when down-sampling the image to
        avoid aliasing artifacts.

    Returns
    -------
    np.ndarray
        The loaded and scaled brain.

    Raises
    ------
    ImageIOLoadException
        If attempt to load a sequence of images with different shapes.
    """
    for i, p in enumerate(
        tqdm(paths_sequence, desc="Loading images", unit="plane")
    ):
        img = tifffile.imread(p)
        if i == 0:
            check_mem(
                img.nbytes * x_scaling_factor * y_scaling_factor,
                len(paths_sequence),
            )
            # TEST: add test case for shape rounding
            volume = np.empty(
                (
                    int(round(img.shape[0] * x_scaling_factor)),
                    int(round(img.shape[1] * y_scaling_factor)),
                    len(paths_sequence),
                ),
                dtype=img.dtype,
            )
        if x_scaling_factor != 1 or y_scaling_factor != 1:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                img = transform.rescale(
                    img,
                    (x_scaling_factor, y_scaling_factor),
                    mode="constant",
                    preserve_range=True,
                    anti_aliasing=anti_aliasing,
                )

        # Raise an error if the shapes of the images aren't the same
        if not volume[:, :, i].shape == img.shape:
            raise ImageIOLoadException("sequence_shape")
        volume[:, :, i] = img
    return volume


def get_size_image_from_file_paths(file_path, file_extension="tif"):
    """
    Returns the size of an image (which is a list of 2D tiff files or a
    single-file tif stack), without loading the whole image.

    Parameters
    ----------
    file_path : str or pathlib.Path
        Filepath of text file containing paths of all 2D files, a
        filepath of a directory containing all 2D files, or a single
        tiff file z-stack.

    file_extension : str, optional
        Optional file extension (if a directory is passed).

    Returns
    -------
    dict
        Dict of image sizes.
    """
    file_path = Path(file_path)
    if file_path.suffix in [".tif", ".tiff"]:
        # read just the metadata
        with tifffile.TiffFile(file_path) as tiff:
            if not len(tiff.series):
                raise ValueError(
                    f"Attempted to load {file_path} but didn't find a z-stack"
                )
            if len(tiff.series) != 1:
                raise ValueError(
                    f"Attempted to load {file_path} but found multiple stacks"
                )

            shape = tiff.series[0].shape
            axes = tiff.series[0].axes.lower()

            if len(shape) != 3:
                raise ValueError(
                    f"Attempted to load {file_path} but didn't find a "
                    f"3-dimensional stack. Found {axes} axes "
                    f"with shape {shape}"
                )
            # axes is e.g. "zxy"
            if set(axes) == {"x", "y", "z"}:
                image_shape = {ax: n for ax, n in zip(axes, shape)}
                return image_shape
            else:  # metadata does not specify axes as expected,
                logging.debug(
                    f"Axis metadata is {axes}, "
                    "which is not the expected set of x,y,z in any order. "
                    "Assume z,y,x"
                )
                image_shape = {
                    ax: n
                    for ax, n in zip(["z", "y", "x"], tiff.series[0].shape)
                }
                return image_shape

    img_paths = get_sorted_file_paths(file_path, file_extension=file_extension)
    z_shape = len(img_paths)

    logging.debug(
        "Loading file: {} to check raw image size" "".format(img_paths[0])
    )
    image_0 = tifffile.imread(img_paths[0])
    y_shape, x_shape = image_0.shape

    image_shape = {"x": x_shape, "y": y_shape, "z": z_shape}
    return image_shape


def get_tiff_meta(
    path: str,
) -> Tuple[Tuple[int, int], np.dtype]:
    with tifffile.TiffFile(path) as tfile:
        nz = len(tfile.pages)
        if not nz:
            raise ValueError(f"tiff file {path} has no pages!")
        first_page = tfile.pages[0]

    return tfile.pages[0].shape, first_page.dtype


lazy_imread = delayed(tifffile.imread)  # lazy reader


def read_z_stack(path):
    """
    Reads z-stack, lazily, if possible.

    If it's a text file or folder with 2D tiff files use dask to read lazily,
    otherwise it's a single file tiff stack and is read into memory.

    :param path: Filename of text file listing 2D tiffs, folder of 2D tiffs,
        or single file tiff z-stack.
    :return: The data as a dask/numpy array.
    """
    if path.endswith(".tiff") or path.endswith(".tif"):
        with tifffile.TiffFile(path) as tiff:
            if not len(tiff.series):
                raise ValueError(
                    f"Attempted to load {path} but couldn't read a z-stack"
                )
            if len(tiff.series) != 1:
                raise ValueError(
                    f"Attempted to load {path} but found multiple stacks"
                )

            axes = tiff.series[0].axes.lower()
            if set(axes) != {"x", "y", "z"}:
                # log that metadata does not specify expected axes
                logging.debug(
                    f"Axis metadata is {axes}, "
                    "which is not the expected set of x,y,z in any order. "
                    "Assume z,y,x"
                )

        return tifffile.imread(path)

    return read_with_dask(path)


def read_with_dask(path):
    """
    Based on https://github.com/tlambert03/napari-ndtiffs
    Reads a folder of tiffs lazily.

    Note that it will make tifffile.imread ignore OME metadata,
    because this can cause issues with correct metadata reading.
    See https://forum.image.sc/t/tifffile-opening-individual-ome-tiff-files-as-single-huge-array-even-when-isolated/77701

    :param path: folder with tifs.
    :return: dask array containing stack of tifs
    """
    path = str(path)
    if path.endswith(".txt"):
        with open(path, "r") as f:
            filenames = [line.rstrip() for line in f.readlines()]

    else:
        filenames = glob.glob(os.path.join(path, "*.tif")) or glob.glob(
            os.path.join(path, "*.tiff")
        )
        if not filenames:
            raise ValueError(
                f"Folder {path} does not contain any .tif or .tiff files"
            )

    shape, dtype = get_tiff_meta(filenames[0])
    lazy_arrays = [
        lazy_imread(fn, is_ome=False)
        for fn in get_sorted_file_paths(filenames)
    ]
    dask_arrays = [
        da.from_delayed(delayed_reader, shape=shape, dtype=dtype)
        for delayed_reader in lazy_arrays
    ]
    stack = da.stack(dask_arrays, axis=0)
    return stack



from pathlib import Path
from typing import Optional

from natsort import natsorted

from brainglobe_utils.general import list


def get_text_lines(
    file: Path,
    return_lines: Optional[int] = None,
    rstrip: Optional[bool] = True,
    sort: Optional[bool] = False,
    remove_empty_lines: Optional[bool] = True,
    encoding: Optional[str] = "utf8",
):
    """
    Return only the nth line of a text file.

    Parameters
    ----------
    file : str or pathlib.Path
        Path to any text file.

    return_lines : int, optional
        Which specific line to read.

    rstrip : bool, optional
        Whether to remove trailing characters.

    sort : bool, optional
        If True, naturally sort the data.

    remove_empty_lines : bool, optional
        If True, ignore empty lines.

    encoding : str, optional
        What encoding the text file has.

    Returns
    -------
    str or list of str
        The nth line or lines.
    """
    with open(file, encoding=encoding) as f:
        lines = f.readlines()
    if rstrip:
        lines = [line.strip() for line in lines]
    if remove_empty_lines:
        lines = list.remove_empty_string(lines)
    if sort:
        lines = natsorted(lines)
    if return_lines is not None:
        lines = lines[return_lines]
    return lines



from typing import Any, Optional

import pandas as pd
from qtpy.QtCore import QAbstractTableModel, QModelIndex, Qt


class DataFrameModel(QAbstractTableModel):
    """
    A Qt table model that wraps a pandas DataFrame for use with Qt
    view widgets.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to be displayed in the Qt view.

    """

    def __init__(self, df: pd.DataFrame):
        """
        Initialize the model with a DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to be displayed.
        """
        super().__init__()
        self._df = df

    def rowCount(self, parent: Optional[QModelIndex] = None) -> int:
        """
        Return the number of rows in the model.

        Parameters
        ----------
        parent : Optional[QModelIndex], optional
            The parent index, by default None.

        Returns
        -------
        int
            The number of rows in the DataFrame.
        """
        return self._df.shape[0]

    def columnCount(self, parent: Optional[QModelIndex] = None) -> int:
        """
        Return the number of columns in the model.

        Parameters
        ----------
        parent : Optional[QModelIndex], optional
            The parent index, by default None.

        Returns
        -------
        int
            The number of columns in the DataFrame.
        """
        return self._df.shape[1]

    def data(
        self, index: QModelIndex, role: int = Qt.DisplayRole
    ) -> Optional[Any]:
        """
        Return the data at the given index for the specified role.

        Parameters
        ----------
        index : QModelIndex
            The index of the data to be retrieved.
        role : int, optional
            The role for which the data is being requested,
            by default Qt.DisplayRole.

        Returns
        -------
        Optional[Any]
            The data at the specified index, or None if the role
            is not Qt.DisplayRole.
        """
        if role == Qt.DisplayRole:
            return str(self._df.iloc[index.row(), index.column()])
        return None

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.DisplayRole,
    ) -> Optional[Any]:
        """
        Return the header data for the specified section and orientation.

        Parameters
        ----------
        section : int
            The section (column or row) for which the header data is requested.
        orientation : Qt.Orientation
            The orientation (horizontal or vertical) of the header.
        role : int, optional
            The role for which the header data is being requested, by
            default Qt.DisplayRole.

        Returns
        -------
        Optional[Any]
            The header data for the specified section and orientation, or
            None if the role is not Qt.DisplayRole.
        """

        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self._df.columns[section]
            if orientation == Qt.Vertical:
                return self._df.index[section]
        return None



import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import Literal
from warnings import warn

from brainglobe_utils.citation.bibtex_fmt import (
    BibTexEntry,
    supported_bibtex_entry_types,
)
from brainglobe_utils.citation.repositories import (
    all_citable_repositories,
    unique_repositories_from_tools,
)
from brainglobe_utils.citation.text_fmt import TextCitation

FORMAT_TO_EXTENSION = {"bibtex": "tex", "text": "txt"}
EXTENSION_TO_FORMAT = {
    value: key for key, value in FORMAT_TO_EXTENSION.items()
}


def cite(
    *tools: str,
    format: Literal["bibtex", "text"] = "text",
    outfile: Path = None,
    cite_software: bool = False,
    newline_separations: int = 2,
    warn_on_unused_info: bool = False,
) -> str:
    """
    Provide citation(s) for the BrainGlobe tool(s) that the user has supplied.

    If two or more aliases point to the same tool, remove duplicates and report
    the duplication has been ignored.

    Parameters
    ----------
    tools: str
        Tool names or aliases that are to be cited.
    format: One of bibtex, default = bibtex
        Reference format to write.
    outfile: str, default = None
        The output file to write to, if provided.
        If None, reference text will be printed to the console.
    cite_software: bool, default = False
        If True, software citations will be preferred over the article
        or journal counterparts, where present in repositories.
        Use if you want to reference the source code of a particular tool,
        rather than acknowledge use of the tool itself.
    newline_separations: int, default = 2
        Number of newline characters to use when separating references, in the
        event that multiple tools are to be cited.
    warn_on_unused_info: bool, default = False
        If True, information parsed from the yaml-content that is not used by
        the citation format will be flagged to the user on output.

    Returns
    -------
    str
        The string of citation text that was produced, in the requested format.

    Raises
    ------
    ValueError
        If the citation data fetched is missing a type specifier, and the
        format requested requires this to be explicitly set.
    """
    unique_repos = unique_repositories_from_tools(
        *tools, report_duplicates=True
    )

    # unique_repos is now a set of all the repositories that we need to cite
    # so we just need to gather all the citations we need
    cite_string = ""

    for repo in unique_repos:
        # Fetch citation information from repository
        citation_info = repo.read_citation_info()

        # Some formats are citation-type agnostic, others are not
        # Attempt to read this key here, and set the value to None
        # if it's not present.
        # This ensures that users don't have to provide the field if
        # they want to cite something that doesn't need the information.
        try:
            citation_type = citation_info["type"]
        except KeyError as e:
            warn(
                f"{repo.name} has no citation type data - "
                "reference may not be generated correctly.\n"
                f"(Caught {str(e)})",
                UserWarning,
            )
            citation_type = None

        # The repo_reference variable will contain the reference string
        repo_reference: str = None

        # Check if there is a preferred-citation field,
        # and whether we want to use that information over the standard
        # software citation.
        # Note that if the alternative citation is also a software
        # citation, it will still be rendered as such.
        if (
            citation_type == "software"
            and not cite_software
            and "preferred-citation" in citation_info.keys()
        ):
            citation_info = citation_info["preferred-citation"]
            citation_type = citation_info["type"]

        # Determine citation format in preparation for writing
        if format == "text":
            # If the user requested the citation sentence,
            # provide this by looking up the expected field.
            reference_instance = TextCitation(
                citation_info, warn_on_not_used=warn_on_unused_info
            )
        else:
            # We need to convert the citation information to
            # a particular format

            # Cite this repository in the desired format
            if format == "bibtex":
                try:
                    citation_class = supported_bibtex_entry_types()[
                        citation_type
                    ]
                except KeyError:
                    raise ValueError(
                        f"Bibtex entries require a supported Bibtex entry type"
                        " to be provided in citation metadata "
                        f"(got {citation_type})"
                    )

                reference_instance: BibTexEntry = citation_class(
                    citation_info,
                    cite_key=f"{repo.name}",
                    warn_on_not_used=warn_on_unused_info,
                )

        # Append the reference to the string we are generating
        repo_reference = reference_instance.generate_ref_string()
        cite_string += f"{repo_reference}" + "\n" * newline_separations

    # Upon looping over each of the repositories, we should be ready to dump
    # the output to the requested location.
    if outfile is not None:
        # Write output to file
        with open(Path(outfile), "w") as output_file:
            output_file.write(cite_string)
    else:
        sys.stdout.write(cite_string)

    return cite_string


class BrainGlobeParser(ArgumentParser):
    """
    Overwrite argparse default behaviour to have usage errors
    print the command-line tool help, rather than throwing the
    error encountered.
    """

    def error(self, msg):
        sys.stderr.write(f"Error: {msg}\nSee usage instructions below:\n")
        self.print_help()
        sys.exit(2)


def cli() -> None:
    """
    Command-line interface for the citation tool.
    """
    parser = BrainGlobeParser(
        description="Citation generation for BrainGlobe tools."
    )

    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List citable BrainGlobe tools, and formats, then exit.",
    )
    parser.add_argument(
        "-s",
        "--software-citations",
        action="store_true",
        help="Explicitly cite software source code over academic papers.",
    )
    parser.add_argument(
        "-w",
        "--warn-unused",
        action="store_true",
        help="Print out when citation information is omitted by "
        "the chosen citation format.",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=str,
        default=None,
        help="Output file to write citations to.",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        default=None,
        help="Citation format to write. "
        "Will be overwritten by the inferred format if the output file "
        "argument is also provided. "
        "Valid formats can be listed with the -l, --list option.",
    )
    parser.add_argument(
        "tools", nargs="*", type=str, help="BrainGlobe tools to be cited."
    )

    arguments = parser.parse_args()

    # Check if we just want to list available options
    if arguments.list:
        # List citable BrainGlobe tools
        sys.stdout.write("Citable BrainGlobe tools by name (source code):\n")
        for repo in sorted(all_citable_repositories()):
            sys.stdout.write(f"\t- {repo.name} ({repo.url})\n")

        # List reference formats
        sys.stdout.write(
            "Available citation formats "
            "(supported file extension), format option:\n"
        )
        sys.stdout.write("\t- BibTex (*.tex), --format bibtex\n")
        sys.stdout.write("\t- Text (*.txt), --format text\n")

        # Terminate
        sys.exit(0)

    # Exit if no tools were requested
    tools_to_cite = arguments.tools
    if not tools_to_cite:
        sys.stderr.write(
            "No tools provided for citation! See usage syntax below:\n"
        )
        parser.print_help()
        sys.exit(1)

    # Pass default values if available
    fmt = getattr(arguments, "format")
    output_file = getattr(arguments, "output_file")
    extension = None

    if output_file is None:
        # With no output file, we cannot infer the format from it.
        # So either use the one that was provided if it is supported,
        # or use the default if it was also omitted
        if fmt is not None and fmt not in FORMAT_TO_EXTENSION:
            raise RuntimeError(f"Output format {fmt} is not supported.")
        elif fmt is None:
            # Use default value as this argument was also not provided
            fmt = "text"
    else:
        # Output file provided - resolve path based on OS.
        output_file = Path(output_file)
        extension = output_file.suffix

        # If there was an extension, then infer the format.
        # This will overwrite the fmt argument.
        if extension in EXTENSION_TO_FORMAT:
            fmt = EXTENSION_TO_FORMAT[extension]
        elif not extension:
            if fmt is None:
                # No extension provided, and no format for the citation
                # provided. Throw error.
                raise RuntimeError(
                    "You have not provided a file extension nor citation "
                    "format to write. You must provide one of these."
                )
            elif fmt not in FORMAT_TO_EXTENSION:
                # Format must be supported to allow writing
                raise RuntimeError(
                    f"{fmt} format is not supported, and your output file "
                    "does not provide an implicit format."
                )
        else:
            # This is an extension that we don't support
            raise RuntimeError(
                f"brainglobe-cite does not support writing {extension} files."
            )

    # Invoke API function
    cite(
        *tools_to_cite,
        format=fmt,
        outfile=output_file,
        cite_software=arguments.software_citations,
        warn_on_unused_info=arguments.warn_unused,
    )
    sys.exit(0)



import psutil
from scipy.ndimage import zoom


class ImageIOLoadException(Exception):
    """
    Custom exception class for errors found loading images with
    brainglobe_utils.IO.image.load

    Alerts the user of: loading a directory containing only a single .tiff,
    loading a single 2D .tiff, loading an image sequence where all 2D images
    don't have the same shape, lack of memory to complete loading.

    Set the error message to self.message to read during testing.
    """

    def __init__(self, error_type=None, total_size=None, free_mem=None):
        if error_type == "single_tiff":
            self.message = (
                "Attempted to load directory containing "
                "a single .tiff file. If the .tiff file "
                "is 3D please pass the full path with "
                "filename. Single 2D .tiff file input is "
                "not supported."
            )

        elif error_type == "2D tiff":
            self.message = "Single 2D .tiff file input is not supported."

        elif error_type == "sequence_shape":
            self.message = (
                "Attempted to load an image sequence where individual 2D "
                "images did not have the same shape. Please ensure all image "
                "files contain the same number of pixels."
            )

        elif error_type == "memory":
            self.message = (
                "Not enough memory on the system to complete "
                "loading operation."
            )
            if total_size is not None and free_mem is not None:
                self.message += (
                    f" Needed {total_size}, only {free_mem} " f"available."
                )

        else:
            self.message = "File failed to load with brainglobe_utils.image."

        super().__init__(self.message)


def check_mem(img_byte_size, n_imgs):
    """
    Checks how much memory is available on the system and compares it to the
    size the stack specified by img_byte_size and n_imgs would take
    once loaded

    Raises an error in case memory is insufficient to load that stack

    Parameters
    ----------
    img_byte_size : int
        The size in bytes of an individual image plane.

    n_imgs : int
        The number of image planes to load.

    Raises
    ------
    ImageIOLoadException
        If not enough memory is available.
    """
    total_size = img_byte_size * n_imgs
    free_mem = psutil.virtual_memory().available
    if total_size >= free_mem:
        raise ImageIOLoadException(
            error_type="memory", total_size=total_size, free_mem=free_mem
        )


def scale_z(volume, scaling_factor):
    """
    Scale the given brain along the z dimension.

    Parameters
    ----------
    volume : np.ndarray
        A brain, typically as a numpy array.

    scaling_factor : float
        The z scaling factor.
    """

    return zoom(volume, (scaling_factor, 1, 1), order=1)



from brainglobe_utils.citation.format import Format


class TextCitation(Format):
    """
    Generates a reference string that can be copy-pasted into a
    text document's bibliography for use as a reference.

    Style of text-based references will be

    <citation-sentence>;
    <authors> (<year>). <title>.
    <journal>, <volume>(<issue>), <doi>, <url>

    Required information:
    - authors
    - title
    - year

    Optional information:
    - citation-sentence
    - journal
    - volume
    - issue
    - doi
    - url
    """

    required = ["authors", "title", "year"]
    optional = [
        "citation-sentence",
        "journal",
        "volume",
        "issue",
        "issn",
        "doi",
        "url",
    ]

    def generate_ref_string(self) -> str:
        """
        Generate a string that encodes the reference, in preparation for
        writing to an output format.
        """
        output_string = ""

        # Include optional sentence if provided
        if getattr(self, "citation_sentence") is not None:
            output_string += f"{getattr(self, 'citation_sentence')};\n"

        # Required fields
        output_string += f"{getattr(self, 'authors')} "
        output_string += f"({getattr(self, 'year')}). "
        output_string += f"{getattr(self, 'title')}.\n"

        # Optional information on final line
        if getattr(self, "journal") is not None:
            output_string += f"{getattr(self, 'journal')}, "
        if getattr(self, "volume") is not None:
            output_string += f"{getattr(self, 'volume')}"
            if getattr(self, "issn") is not None:
                output_string += f"({getattr(self, 'issn')}), "
            elif getattr(self, "issue") is not None:
                output_string += f"({getattr(self, 'issue')}), "
        if getattr(self, "doi") is not None:
            output_string += f"{getattr(self, 'doi')}, "
        if getattr(self, "url") is not None:
            output_string += f"{getattr(self, 'url')}, "

        # Trim hanging comma if present
        output_string = output_string.strip()
        output_string = output_string.strip(",")

        return output_string



from importlib.resources import files
from typing import Optional

from qtpy.QtWidgets import QGroupBox, QHBoxLayout, QLabel, QWidget


def _docs_links_widget(
    package_name: str,
    package_tagline: str,
    tutorial_file_name: Optional[str] = None,
    documentation_path: Optional[str] = None,
    citation_doi: Optional[str] = None,
    github_repo_name: Optional[str] = None,
    help_text: Optional[str] = None,
    parent: Optional[QWidget] = None,
):
    lines = [
        "<h3>",
        f"<p>{package_tagline}</p>",
        "<p><a href='https://brainglobe.info' style='color:gray;'>"
        "Website</a></p>",
    ]

    if tutorial_file_name:
        lines.append(
            f"<p>"
            f"<a href='https://brainglobe.info/tutorials/{tutorial_file_name}'"
            f" style='color:gray;'>Tutorial</a></p>"
        )

    if documentation_path:
        lines.append(
            f"<p><a href="
            f"'https://brainglobe.info/documentation/{documentation_path}' "
            f"style='color:gray;'>Documentation</a></p>"
        )

    if github_repo_name is None:
        github_repo_name = package_name
    lines.append(
        f"<p><a href='https://github.com/brainglobe/{github_repo_name}' "
        f"style='color:gray;'>Source</a></p>"
    )

    if citation_doi:
        lines.append(
            f"<p><a href='{citation_doi}' style='color:gray;'>Citation</a></p>"
        )

    if help_text:
        lines.append(f"<p><small>{help_text}</small></p>")

    lines.append("</h3>")
    docs_links_widget = QLabel("\n".join(lines), parent=parent)
    docs_links_widget.setOpenExternalLinks(True)
    return docs_links_widget


def _logo_widget(package_name: str, parent: Optional[QWidget] = None):
    brainglobe_logo = files("brainglobe_utils").joinpath("qtpy/brainglobe.png")

    _logo_html = f"""
    <h1>
    <img src="{brainglobe_logo}"width="100">
    <p>{package_name}</p>
    </h1>
    """  # noqa W605

    return QLabel(_logo_html, parent=None)


def header_widget(
    package_name: str,
    package_tagline: str,
    tutorial_file_name: Optional[str] = None,
    documentation_path: Optional[str] = None,
    citation_doi: Optional[str] = None,
    github_repo_name: Optional[str] = None,
    help_text: Optional[str] = None,
    parent: Optional[QWidget] = None,
) -> QGroupBox:
    """
    Render HTML in a QGroupBox with a BrainGlobe logo and links to the docs
     and source code.

    Parameters
    ----------
    package_name : str
        The name of the package, e.g. "brainrender-napari"
    package_tagline : str
        The tagline for the package
    tutorial_file_name : str, optional
        The name of the tutorial file (must include .html),
         e.g. "brainrender-qtpy.html"
    documentation_path : str, optional
        Path of documentation file relative to
        https://brainglobe.info/documentation/ (must include .html),
        e.g. "cellfinder/user-guide/napari-plugin/index.html"
    citation_doi : str, optional
        Doi of citation e.g. "https://doi.org/10.1371/journal.pcbi.1009074"
    github_repo_name : str, optional
        Name of github repository inside the BrainGlobe organisation
        (https://github.com/brainglobe). Only necessary if the github
        repository name isn't the same as the package name.
    help_text : str, optional
        Help text to display at the bottom or the header e.g.
        "For help, hover the cursor over each parameter."
    parent : QWidget, optional
        The parent widget, defaults to None

    Returns
    -------
    QGroupBox
        A QGroupBox with the rendered HTML containing the logo and links
    """

    box = QGroupBox(parent)
    box.setFlat(True)
    box.setLayout(QHBoxLayout())
    box.layout().addWidget(_logo_widget(package_name, parent=box))
    box.layout().addWidget(
        _docs_links_widget(
            package_name,
            package_tagline,
            tutorial_file_name,
            documentation_path,
            citation_doi,
            github_repo_name,
            help_text,
            parent=box,
        )
    )

    return box



import glob
import logging
import os
import platform
import shutil
import subprocess
from pathlib import Path
from tempfile import gettempdir
from typing import Union

import psutil
from natsort import natsorted
from slurmio import slurmio
from tqdm import tqdm

from brainglobe_utils.general.exceptions import CommandLineInputError
from brainglobe_utils.general.string import get_text_lines

# On Windows, max_workers must be less than or equal to 61
# https://docs.python.org/3/library/concurrent.futures.html#processpoolexecutor
MAX_PROCESSES_WINDOWS = 61


def ensure_extension(
    file_path: Union[str, os.PathLike], extension: str
) -> Path:
    """
    Ensure that the given file path has the specified extension.

    If the file path does not already have the specified extension,
    it changes the file path to have that extension.

    Parameters
    ----------
    file_path : Union[str, os.PathLike]
        The path to the file.
    extension : str
        The desired file extension (should include the dot, e.g., '.txt').

    Returns
    -------
    Path
        The Path object with the ensured extension.
    """
    path = Path(file_path)
    if path.suffix != extension:
        path = path.with_suffix(extension)
    return path


def replace_extension(file, new_extension, check_leading_period=True):
    """
    Replaces the file extension of a given file.

    Parameters
    ----------
    file : str
        Input file path with file extension to replace.

    new_extension : str
        New file extension.

    check_leading_period : bool, optional
        If True, any leading period of the new extension is removed,
        preventing "file..txt".

    Returns
    -------
    str
        File path with new file extension.
    """
    if check_leading_period:
        new_extension = remove_leading_character(new_extension, ".")
    return os.path.splitext(file)[0] + "." + new_extension


def remove_leading_character(string, character):
    """
    If "string" starts with "character", strip that leading character away.
    Only removes the first instance.

    Parameters
    ----------
    string : str
        Input string.

    character : str
        Character to be stripped if found at the beginning of the string.

    Returns
    -------
    str
        String without the specified leading character.
    """
    if string.startswith(character):
        return string[1:]
    else:
        return string


def ensure_directory_exists(directory):
    """
    If a directory doesn't exist, make it. Works for pathlib objects,
    and strings.

    Parameters
    ----------
    directory : str or pathlib.Path
        Directory to be created if it doesn't exist.
    """
    if isinstance(directory, str):
        if not os.path.exists(directory):
            os.makedirs(directory)
    elif isinstance(directory, Path):
        directory.mkdir(exist_ok=True)


def get_sorted_file_paths(file_path, file_extension=None, encoding=None):
    """
    Sorts file paths with numbers "naturally" (i.e. 1, 2, 10, a, b), not
    lexicographically (i.e. 1, 10, 2, a, b).

    Parameters
    ----------
    file_path : list of str or str or pathlib.Path
        List of file paths, or path of a text file containing these paths,
        or path of a directory containing files.

    file_extension : str, optional
        Only return filepaths with this extension (if a directory is passed).

    encoding : str, optional
        If opening a text file, what encoding it has.
        Default is None (platform dependent).

    Returns
    -------
    list of str
        Sorted list of file paths.
    """
    if isinstance(file_path, list):
        return natsorted(file_path)

    # assume if not a list, is a file path
    file_path = Path(file_path)
    if file_path.suffix == ".txt":
        return get_text_lines(file_path, sort=True, encoding=encoding)
    elif file_path.is_dir():
        if file_extension is None:
            file_path = glob.glob(os.path.join(file_path, "*"))
        else:
            file_path = glob.glob(
                os.path.join(file_path, "*" + file_extension)
            )
        return natsorted(file_path)

    else:
        message = (
            "Input file path is not a recognised format. Please check it "
            "is a list of file paths, a text file of these paths, or a "
            "directory containing image files."
        )
        raise NotImplementedError(message)


def check_path_in_dir(file_path, directory_path):
    """
    Check if a file path is in a directory.

    Parameters
    ----------
    file_path : str or pathlib.Path
        Full path to a file.

    directory_path : str or pathlib.Path
        Full path to a directory the file may be in.

    Returns
    -------
    bool
        True if the file is in the directory, False otherwise.
    """
    directory = Path(directory_path)
    parent = Path(file_path).parent
    return parent == directory


def get_num_processes(
    min_free_cpu_cores=2,
    ram_needed_per_process=None,
    fraction_free_ram=0.1,
    n_max_processes=None,
    max_ram_usage=None,
    enforce_single_cpu=True,
):
    """
    Determine how many CPU cores to use, based on a minimum number of CPU cores
    to leave free, and an optional maximum number of processes.

    Cluster computing aware for the SLURM job scheduler, and not yet
    implemented for other environments.

    Parameters
    ----------
    min_free_cpu_cores : int, optional
        How many CPU cores to leave free.

    ram_needed_per_process : float, optional
        Memory requirements per process. Set this to ensure that the number of
        processes isn't too high.

    fraction_free_ram : float, optional
        Fraction of the RAM to ensure stays free regardless of the current
        program.

    n_max_processes : int, optional
        Maximum number of processes.

    max_ram_usage : float, optional
        Maximum amount of RAM (in bytes) to use (although available may be
        lower).

    enforce_single_cpu : bool, optional
        Ensure that >=1 CPU core is chosen.

    Returns
    -------
    int
        Number of processes to use.
    """
    logging.debug("Determining the maximum number of CPU cores to use")

    n_cpu_cores = get_cores_available()
    n_cpu_cores = n_cpu_cores - min_free_cpu_cores

    logging.debug(f"Number of CPU cores available is: {n_cpu_cores}")

    if ram_needed_per_process is not None:
        n_processes = limit_cores_based_on_memory(
            n_cpu_cores,
            ram_needed_per_process,
            fraction_free_ram,
            max_ram_usage,
        )
    else:
        n_processes = n_cpu_cores

    n_max_processes = limit_cpus_windows(n_max_processes)

    if n_max_processes is not None:
        if n_max_processes < n_processes:
            logging.debug(
                f"Limiting the number of processes to {n_max_processes} based"
                f" on other considerations."
            )
        n_processes = min(n_processes, n_max_processes)

    if enforce_single_cpu:
        if n_processes < 1:
            logging.debug("Forcing number of processes to be 1")
            n_processes = 1

    logging.debug(f"Setting number of processes to: {n_processes}")
    return int(n_processes)


def limit_cpus_windows(n_max_processes):
    if platform.system() == "Windows":
        if n_max_processes is not None:
            n_max_processes = min(n_max_processes, MAX_PROCESSES_WINDOWS)
    return n_max_processes


def get_cores_available():
    try:
        os.environ["SLURM_JOB_ID"]
        n_cpu_cores = slurmio.SlurmJobParameters().allocated_cores
    except KeyError:
        n_cpu_cores = psutil.cpu_count()

    return n_cpu_cores


def limit_cores_based_on_memory(
    n_cpu_cores, ram_needed_per_process, fraction_free_ram, max_ram_usage
):
    cores_w_sufficient_ram = how_many_cores_with_sufficient_ram(
        ram_needed_per_process,
        fraction_free_ram=fraction_free_ram,
        max_ram_usage=max_ram_usage,
    )
    n_processes = min(n_cpu_cores, cores_w_sufficient_ram)
    logging.debug(
        f"Based on memory requirements, up to {cores_w_sufficient_ram} "
        f"cores could be used. Therefore setting the number of "
        f"processes to {n_processes}."
    )
    return n_processes


def how_many_cores_with_sufficient_ram(
    ram_needed_per_cpu, fraction_free_ram=0.1, max_ram_usage=None
):
    """
    Based on the amount of RAM needed per CPU core for a multiprocessing task,
    work out how many CPU cores could theoretically be used based on the
    amount of free RAM. N.B. this does not relate to how many CPU cores
    are actually available.

    Parameters
    ----------
    ram_needed_per_cpu : float
        Memory requirements per process. Set this to ensure that the number of
        processes isn't too high.

    fraction_free_ram : float, optional
        Fraction of the RAM to ensure stays free regardless of the current
        program.

    max_ram_usage : float, optional
        The Maximum amount of RAM (in bytes) to use (although available may be
        lower).

    Returns
    -------
    int
        How many CPU cores could be theoretically used based on the amount of
        free RAM.
    """

    try:
        # if in slurm environment
        os.environ["SLURM_JOB_ID"]
        # Only allocated memory (not free). Assumes that nothing else will be
        # running
        free_mem = slurmio.SlurmJobParameters().allocated_memory
    except KeyError:
        free_mem = get_free_ram()

    logging.debug(f"Free memory is: {free_mem} bytes.")

    if max_ram_usage is not None:
        free_mem = min(free_mem, max_ram_usage)
        logging.debug(
            f"Maximum memory has been set as: {max_ram_usage} "
            f"bytes, so using: {free_mem} as the maximum "
            f"available memory"
        )

    free_mem = free_mem * (1 - fraction_free_ram)
    cores_w_sufficient_ram = free_mem / ram_needed_per_cpu
    return int(cores_w_sufficient_ram // 1)


def disk_free_gb(file_path):
    """
    Return the free disk space, on a disk defined by a file path.

    Parameters
    ----------
    file_path : str
        File path on the disk to be checked.

    Returns
    -------
    float
        Free space in GB.
    """
    if platform.system() == "Windows":
        drive, _ = os.path.splitdrive(file_path)
        total, used, free = shutil.disk_usage(drive)
        return free / 1024**3
    else:
        stats = os.statvfs(file_path)
        return (stats.f_frsize * stats.f_bavail) / 1024**3


def get_free_ram():
    """
    Returns the amount of free RAM in bytes.

    Returns
    -------
    int
        Available RAM in bytes.
    """
    return psutil.virtual_memory().available


def safe_execute_command(cmd, log_file_path=None, error_file_path=None):
    """
    Executes a command in the terminal, making sure that the output can
    be logged even if execution fails during the call.

    From https://github.com/SainsburyWellcomeCentre/amap_python by
    Charly Rousseau (https://github.com/crousseau).

    Parameters
    ----------
    cmd : str
        Command to be executed.

    log_file_path : str, optional
        File path to log the output.

    error_file_path : str, optional
        File path to log any errors.
    """
    if log_file_path is None:
        log_file_path = os.path.abspath(
            os.path.join(gettempdir(), "safe_execute_command.log")
        )
    if error_file_path is None:
        error_file_path = os.path.abspath(
            os.path.join(gettempdir(), "safe_execute_command.err")
        )

    with (
        open(log_file_path, "w") as log_file,
        open(error_file_path, "w") as error_file,
    ):
        try:
            subprocess.check_call(
                cmd, stdout=log_file, stderr=error_file, shell=True
            )
        except subprocess.CalledProcessError:
            hline = "-" * 25
            try:
                with open(error_file_path, "r") as err_file:
                    errors = err_file.readlines()
                    errors = "".join(errors)
                with open(log_file_path, "r") as _log_file:
                    logs = _log_file.readlines()
                    logs = "".join(logs)
                raise SafeExecuteCommandError(
                    "\n{0}\nProcess failed:\n {1}"
                    "{0}\n"
                    "{2}\n"
                    "{0}\n"
                    "please read the logs at {3} and {4}\n"
                    "{0}\n"
                    "command: {5}\n"
                    "{0}".format(
                        hline,
                        errors,
                        logs,
                        log_file_path,
                        error_file_path,
                        cmd,
                    )
                )
            except IOError as err:
                raise SafeExecuteCommandError(
                    f"Process failed: please read the logs at {log_file_path} "
                    f"and {error_file_path}; command: {cmd}; err: {err}"
                )


class SafeExecuteCommandError(Exception):
    pass


def delete_directory_contents(directory, progress=False):
    """
    Removes all contents of a directory.

    Parameters
    ----------
    directory : str
        Directory with files to be removed.

    progress : bool, optional
        Whether to show a progress bar.
    """
    files = os.listdir(directory)
    if progress:
        for f in tqdm(files):
            os.remove(os.path.join(directory, f))
    else:
        for f in files:
            os.remove(os.path.join(directory, f))


def check_path_exists(file):
    """
    Returns True if a file exists, otherwise throws a FileNotFoundError.

    Parameters
    ----------
    file : str or pathlib.Path
        Input file.

    Returns
    -------
    bool
        True if the file exists.

    Raises
    ------
    FileNotFoundError
        If the file doesn't exist.
    """
    file = Path(file)
    if file.exists():
        return True
    else:
        raise FileNotFoundError


def catch_input_file_error(path):
    """
    Catches if an input path doesn't exist, and returns an informative error.

    Parameters
    ----------
    path : str or pathlib.Path
        Input file path.

    Raises
    ------
    CommandLineInputError
        If the file doesn't exist.
    """
    try:
        check_path_exists(path)
    except FileNotFoundError:
        message = (
            "File path: '{}' cannot be found. Please check your input "
            "arguments.".format(path)
        )
        raise CommandLineInputError(message)



from typing import List, Optional

from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QVBoxLayout, QWidget
from superqt.collapsible import QCollapsible


class CollapsibleWidget(QCollapsible):
    """
    Custom collapsible widget.

    Attributes
    ----------
    toggled_signal_with_self : Signal
        Emitted when the CollapsibleWidget is toggled.
        Provides the CollapsibleWidget instance and the new state.

    Parameters
    ----------
    title : str, optional
        The title of the CollapsibleWidget.
    parent : QWidget or None, optional
        The parent widget.
    """

    toggled_signal_with_self = Signal(QCollapsible, bool)

    def __init__(
        self,
        title: str = "",
        parent: Optional[QWidget] = None,
        expanded_icon: Optional[str] = "▼",
        collapsed_icon: Optional[str] = "▶",
    ):
        """
        Initializes a new CollapsibleWidget instance.

        Parameters
        ----------
        title : str, optional
            The title of the CollapsibleWidget.
        parent : QWidget or None, optional
            The parent widget.
        """
        super().__init__(
            title,
            parent,
            expandedIcon=expanded_icon,
            collapsedIcon=collapsed_icon,
        )
        self.currently_expanded = False

        self.toggled.connect(self._on_toggle)

    def _on_toggle(self, state):
        """
        Handles the toggled signal by emitting the custom signal
         with the CollapsibleWidget instance and the new state.

        Parameters
        ----------
        state : bool
            The new state of the CollapsibleWidget
            (True if expanded, False if collapsed).
        """
        self.toggled_signal_with_self.emit(self, state)


class CollapsibleWidgetContainer(QWidget):
    """
    Container for multiple CollapsibleWidgets with the ability to add,
    remove, and synchronize their states. Non-CollapsibleWidgets can also
    be added.

    Methods
    -------
    add_widget(QWidget)
        Adds a widget to the CollapsibleWidgetContainer.
    remove_drawer(QWidget or CollapsibleWidget)
        Removes a widget from the CollapsibleWidgetContainer.
    _update_drawers(signalling_drawer, state)
        Private method to synchronize drawer states.
    """

    def __init__(self, parent=None):
        """
        Initializes a new CollapsibleWidgetContainer instance.
        """
        super().__init__(parent)
        self.setLayout(QVBoxLayout())

        self.layout().setAlignment(Qt.AlignTop)
        self.layout().setSpacing(0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.collapsible_widgets: List[CollapsibleWidget] = []

    def add_widget(
        self, widget: QWidget, collapsible: bool = True, widget_title: str = ""
    ):
        """
        Adds a QWidget or a CollapsibleWidget to the chest.

        Parameters
        ----------
        widget : QWidget
            The widget instance to be added.
        collapsible : bool, optional
            Whether the widget should be collapsible.
        widget_title : str, optional
            The title of the widget.
        """
        if collapsible:
            collapsible_widget = CollapsibleWidget(widget_title, parent=self)
            collapsible_widget.setContent(widget)
            collapsible_widget.toggled_signal_with_self.connect(
                self._update_drawers
            )
            collapsible_widget.collapse(animate=False)
            self.collapsible_widgets.append(collapsible_widget)
            self.layout().addWidget(collapsible_widget, 0, Qt.AlignTop)
        else:
            self.layout().addWidget(widget, 0, Qt.AlignTop)

    def remove_widget(self, widget: QWidget):
        """
        Removes a widget from the chest.

        Parameters
        ----------
        widget : QWidget
            The widget instance to be removed.

        Raises
        ------
        ValueError
            If the widget is not found.
        """
        for i in range(self.layout().count()):
            child_widget = self.layout().itemAt(i).widget()
            if (
                isinstance(child_widget, CollapsibleWidget)
                and child_widget.content() is widget
            ):
                self.layout().removeWidget(child_widget)
                self.collapsible_widgets.remove(child_widget)
                child_widget.toggled_signal_with_self.disconnect(
                    self._update_drawers
                )
                return
            elif child_widget is widget:
                self.layout().removeWidget(widget)
                return

        raise ValueError("Widget not found")

    def _update_drawers(
        self, signalling_widget: CollapsibleWidget, state: bool
    ):
        """
        Synchronizes CollapsibleWidget states to ensure only one
        CollapsibleWidget is expanded at a time.

        Parameters
        ----------
        signalling_widget : CollapsibleWidget
            The CollapsibleWidget emitting the signal.
        state : bool
            The new state of the signalling_widget.
        """
        if state:
            for collapsible_widget in self.collapsible_widgets:
                if collapsible_widget is not signalling_widget:
                    collapsible_widget.collapse(False)



from typing import Dict

import requests
from yaml import safe_load

BASE_URL = "https://raw.githubusercontent.com"


def fetch_from_github(
    user: str,
    repo: str,
    file: str = "CITATION.cff",
    branch: str = "main",
) -> requests.Response:
    """
    Fetches the content of a file hosted on GitHub,
    returning the request object.

    Content is fetched from user/repository/file on github.com.

    Parameters
    ----------
    user : str
        Username or organisation the repository can be found under.
    repository : str
        Name of the repository.
    file : str, default = CITATION.cff
        Path to the file from the repository root.
    branch : str, default = main
        Branch to fetch the file from.

    Returns
    -------
    requests.Response
        The response containing the file contents as text.

    Raises
    ------
    RuntimeError
        If the request returns with a bad status code,
        indicating fetch failure.
    """
    url = f"{BASE_URL}/{user}/{repo}/{branch}/{file}"

    r = requests.get(url)

    if not r.ok:
        raise RuntimeError(
            "Bad request or response, got status code "
            f"{r.status_code} when fetching from {url}"
        )

    return r


def yaml_str_to_dict(text: str) -> Dict[str, str]:
    """
    Cast a string of text in yaml syntax to a Python dictionary.
    """
    return safe_load(text)



"""
Based on https://github.com/SainsburyWellcomeCentre/niftynet_cell_count by
Christian Niedworok (https://github.com/cniedwor).
"""

import math
import os
import re
import threading
from collections import defaultdict
from functools import total_ordering
from typing import (
    Any,
    DefaultDict,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from xml.etree import ElementTree
from xml.etree.ElementTree import Element as EtElement

import numpy as np
from numba import njit, objmode
from tqdm import tqdm


@total_ordering
class Cell:
    """
    A class representing a cell with a specific type.

    Parameters
    ----------
    pos : str or ElementTree.Element or Dict[str, float] or List of float
        Cell position (x, y, z). Input can be a filename containing the x/y/z
        position, an xml marker element, a dictionary with keys "x", "y" and
        "z", or a list of the positions [x, y, z].

    cell_type : int or str or None
        Cell type represented by an integer: 1 for unknown/no cell,
        2 for cell, -1 for artifact. The usual way to set this is to use:
        Cell.ARTIFACT, Cell.CELL, Cell.UNKNOWN, or Cell.NO_CELL as input.
        You can also pass "cell" or "no_cell", as well as None (which will
        map to Cell.UNKNOWN).

    Attributes
    ----------
    x : float
        X position.

    y : float
        Y position.

    z : float
        Z position.

    type : int
        Cell type. 1 for unknown/no cell, 2 for cell, -1 for artifact.

    transformed_x : float
        Transformed x position.

    transformed_y : float
        Transformed y position.

    transformed_z : float
        Transformed z position.

    structure_id : int
        ID of brain structure.

    hemisphere : str
        Hemisphere of brain.
    """

    # integers for self.type
    ARTIFACT = -1
    CELL = 2
    UNKNOWN = 1

    # for classification compatibility
    NO_CELL = 1

    def __init__(
        self,
        pos: Union[str, ElementTree.Element, Dict[str, float], List[float]],
        cell_type: int,
    ):
        if isinstance(pos, str):
            pos = pos_from_file_name(os.path.basename(pos))
        if isinstance(pos, ElementTree.Element):
            pos = pos_from_xml_marker(pos)
        if isinstance(pos, dict):
            pos = pos_from_dict(pos)
        pos = self._sanitize_position(pos)
        x, y, z = [int(p) for p in pos]
        self.x: float = x
        self.y: float = y
        self.z: float = z

        self.transformed_x: float = x
        self.transformed_y: float = y
        self.transformed_z: float = z

        self.structure_id = None
        self.hemisphere = None

        self.type: int
        if cell_type is None:
            self.type = Cell.UNKNOWN
        elif str(cell_type).lower() == "cell":
            self.type = Cell.CELL
        elif str(cell_type).lower() == "no_cell":
            self.type = Cell.ARTIFACT
        else:
            self.type = int(cell_type)

    def _sanitize_position(
        self, pos: List[float], verbose: bool = True
    ) -> List[float]:
        out = []
        for coord in pos:
            if math.isnan(coord):
                if verbose:
                    print(
                        "WARNING: NaN position for for cell\n"
                        "defaulting to 1"
                    )
                coord = 1
            out.append(coord)
        return out

    def _transform(
        self,
        x_scale: float = 1.0,
        y_scale: float = 1.0,
        z_scale: float = 1.0,
        x_offset: float = 0,
        y_offset: float = 0,
        z_offset: float = 0,
        integer: bool = False,
    ) -> Tuple[float, float, float]:
        x = self.x
        y = self.y
        z = self.z

        x += x_offset
        y += y_offset
        z += z_offset

        x *= x_scale
        y *= y_scale
        z *= z_scale

        if integer:
            return int(round(x)), int(round(y)), int(round(z))
        else:
            return x, y, z

    def transform(
        self,
        x_scale: float = 1.0,
        y_scale: float = 1.0,
        z_scale: float = 1.0,
        x_offset: float = 0,
        y_offset: float = 0,
        z_offset: float = 0,
        integer: bool = False,
    ) -> None:
        """
        Scale and/or offset the cell position. .x / .y / .z will be updated.

        Parameters
        ----------
        x_scale : float
            Value to scale the x coordinate by.

        y_scale : float
            Value to scale the y coordinate by.

        z_scale : float
            Value to scale the z coordinate by.

        x_offset : float
            Distance to offset in x.

        y_offset : float
            Distance to offset in y.

        z_offset : float
            Distance to offset in z.

        integer : bool
            Whether to round xyz position to nearest integer.
        """
        transformed_coords = self._transform(
            x_scale, y_scale, z_scale, x_offset, y_offset, z_offset, integer
        )
        self.x, self.y, self.z = transformed_coords

    def soft_transform(
        self,
        x_scale: float = 1.0,
        y_scale: float = 1.0,
        z_scale: float = 1.0,
        x_offset: float = 0,
        y_offset: float = 0,
        z_offset: float = 0,
        integer: bool = False,
    ) -> None:
        """
        Scale and/or offset the cell position. New values will be saved into
        .transformed_x, .transformed_y and .transformed_z. Original .x, .y and
        .z will remain unchanged.

        See Also
        --------
        Cell.transform : For description of parameters.
        """
        transformed_coords = self._transform(
            x_scale, y_scale, z_scale, x_offset, y_offset, z_offset, integer
        )
        (
            self.transformed_x,
            self.transformed_y,
            self.transformed_z,
        ) = transformed_coords

    def flip_x_y(self) -> None:
        """Swap the x and y coordinate"""
        self.y, self.x = self.x, self.y

    def is_cell(self) -> bool:
        return self.type == Cell.CELL

    def to_xml_element(self) -> EtElement:
        """
        Create an xml element representing the cell, including its xyz
        coordinate.
        """
        sub_elements = [EtElement("Marker{}".format(axis)) for axis in "XYZ"]
        coords = [int(coord) for coord in (self.x, self.y, self.z)]
        for sub_element, coord in zip(sub_elements, coords):
            if coord < 1:
                print(
                    "WARNING: negative coordinate found at {}\n"
                    "defaulting to 1".format(coord)
                )
                coord = 1  # FIXME:
            sub_element.text = str(coord)

        element = EtElement("Marker")
        element.extend(sub_elements)
        return element

    def __eq__(self, other: Any) -> bool:
        """Return true if position and type of the cells are equal"""
        if not isinstance(other, self.__class__):
            return False
        return (self.x, self.y, self.z, self.type) == (
            other.x,
            other.y,
            other.z,
            other.type,
        )

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __lt__(self, other: Any) -> Union[bool, NotImplementedError]:
        if self == other:
            return False
        try:
            if self.z < other.z:
                return True
            elif self.z > other.z:
                return False
            elif self.y < other.y:
                return True
            elif self.y > other.y:
                return False
            elif self.x < other.x:
                return True
            else:
                return False
        except AttributeError as err:
            return NotImplementedError(
                "comparison to {} is not implemented, {}".format(
                    type(other), err
                )
            )

    def __str__(self) -> str:
        return "Cell: x: {}, y: {}, z: {}, type: {}".format(
            int(self.x), int(self.y), int(self.z), self.type
        )

    def __repr__(self) -> str:
        return "{}, ({}, {})".format(
            self.__class__, [self.x, self.y, self.z], self.type
        )

    def to_dict(self) -> Dict[str, float]:
        return {"x": self.x, "y": self.y, "z": self.z, "type": self.type}

    def __hash__(self) -> int:
        return hash(str(self))


class UntypedCell(Cell):
    """
    A class representing a cell with no type.

    Parameters
    ----------
    pos : str or ElementTree.Element or Dict[str, float] or List of float
        Cell position (x, y, z). Input can be a filename containing the x/y/z
        position, an xml marker element, a dictionary with keys "x", "y" and
        "z", or a list of the positions [x, y, z].

    See Also
    --------
    Cell : For description of attributes.
    """

    def __init__(
        self,
        pos: Union[str, ElementTree.Element, Dict[str, float], List[float]],
    ) -> None:
        super(UntypedCell, self).__init__(pos, self.UNKNOWN)

    @property
    def type(self) -> int:
        return self.UNKNOWN

    @type.setter
    def type(self, value: int) -> None:
        pass

    @classmethod
    def from_cell(cls, cell: Cell) -> "UntypedCell":
        return cls([cell.x, cell.y, cell.z])

    def to_cell(self) -> Cell:
        return Cell([self.x, self.y, self.z], self.type)


def pos_from_dict(position_dict: Dict[str, float]) -> List[float]:
    """Return [x, y, z] position from dictionary with keys of x, y and z"""
    return [position_dict["x"], position_dict["y"], position_dict["z"]]


def pos_from_xml_marker(element: ElementTree.Element) -> List[float]:
    """Return [x, y, z] position from xml marker"""
    marker_names = ["Marker{}".format(axis) for axis in "XYZ"]
    markers = [element.find(marker_name) for marker_name in marker_names]
    pos = [marker.text for marker in markers if marker is not None]
    return [float(num) for num in pos if num is not None]


def pos_from_file_name(file_name: str) -> List[float]:
    """Return [x, y, z] position from filename. For example,
    'pCellz10y522x392Ch0.tif' would return [392, 522, 10]"""
    x = re.findall(r"x\d+", file_name.lower())
    y = re.findall(r"y\d+", file_name.lower())
    z = re.findall(r"z\d+", file_name.lower())
    return [int(p) for p in (x[-1][1:], y[-1][1:], z[-1][1:])]


def group_cells_by_z(cells: List[Cell]) -> DefaultDict[float, List[Cell]]:
    """
    For a list of Cells return a dict of lists of cells, grouped by plane.

    Parameters
    ----------
    cells : List of Cell
        List of cells from brainglobe_utils.cells.cells.Cell

    Returns
    -------
    DefaultDict
        defaultdict, with each key being a plane (e.g. 1280) and each entry
        being a list of Cells
    """
    cells_groups = defaultdict(list)
    for cell in cells:
        cells_groups[cell.z].append(cell)
    return cells_groups


class MissingCellsError(Exception):
    """Custom exception class for when no cells are found in a file"""

    pass


def to_numpy_pos(
    cells: List[Cell], cell_type: Optional[int] = None
) -> np.ndarray:
    """
    Takes a list of Cell objects, selects only cells of type `cell_type` (if
    not None) and returns a single 2d array of shape Nx3 with the
    positions of the cells.
    """
    # for large cell list, pre-compute size
    n = len(cells)
    if cell_type is not None:
        n = sum([cell.type == cell_type for cell in cells])
    np_cells = np.empty((n, 3), dtype=np.float64)

    i = 0
    for cell in cells:
        if cell_type is not None and cell.type != cell_type:
            continue
        np_cells[i, :] = cell.x, cell.y, cell.z
        i += 1

    return np_cells


def from_numpy_pos(pos: np.ndarray, cell_type: int) -> List[Cell]:
    """
    Takes a 2d numpy position array of shape Nx3 and returns a list of Cell
    objects of given cell_type from those positions.
    """
    cells = []
    for i in range(pos.shape[0]):
        cell = Cell(pos=pos[i, :].tolist(), cell_type=cell_type)
        cells.append(cell)

    return cells


def match_cells(
    cells: List[Cell],
    other: List[Cell],
    threshold: float = np.inf,
    pre_match: bool = True,
) -> Tuple[List[int], List[Tuple[int, int]], List[int]]:
    """
    Given two lists of cells. It finds a pairing of cells from `cells` and
    `other` such that the distance (euclidian) between the assigned matches
    across all `cells` is minimized.

    Remaining cells (e.g. if one list is longer or if there are matches
    violating the threshold) are indicated as well.

    E.g.::

        >>> cells = [
        >>>     Cell([20, 20, 20], Cell.UNKNOWN),
        >>>     Cell([10, 10, 10], Cell.UNKNOWN),
        >>>     Cell([40, 40, 40], Cell.UNKNOWN),
        >>>     Cell([50, 50, 50], Cell.UNKNOWN),
        >>> ]
        >>> other = [
        >>>     Cell([5, 5, 5], Cell.UNKNOWN),
        >>>     Cell([15, 15, 15], Cell.UNKNOWN),
        >>>     Cell([35, 35, 35], Cell.UNKNOWN),
        >>>     Cell([100, 100, 100], Cell.UNKNOWN),
        >>>     Cell([200, 200, 200], Cell.UNKNOWN),
        >>> ]
        >>> match_cells(cells, other, threshold=20)
        ([3], [[0, 1], [1, 0], [2, 2]], [3, 4])

    Parameters
    ----------
    cells : list of Cells.
    other : Another list of Cells.
    threshold : float, optional. Defaults to np.inf.
        The threshold to use to remove bad matches. Any match pair whose
        distance is greater than the threshold will be exluded from the
        matching.
    pre_match : bool, optional. Defaults to True.
        If True, we will (internally) first efficiently find all the pairs of
        `cells` and `others` which are each at the same position in space. Then
        we run the optimization to find the best matching on the remaining.

        This will significantly speed up the matching, if there are pairs of
        cells on top of each other in each set.

    Returns
    -------
    tuple :
        missing_cells: List of all the indices of `cells` that found no match
            in `other` (sorted).
        good_matches: List of tuples with all the (cells, other) indices pairs
            that matched below the threshold. It's sorted by the `cells`
            column.
        missing_other: List of all the indices of `other` that found no match
            in `cells` (sorted).
    """
    if __progress_update.updater is not None:
        # I can't think of an instance where this will happen, but better safe
        raise TypeError(
            "An instance of match_cells is already running in this "
            "thread. Try running again once it completes"
        )
    c1 = to_numpy_pos(cells)
    c2 = to_numpy_pos(other)

    # c1 must be smaller or equal in length than c2
    flip = len(cells) > len(other)
    if flip:
        c1, c2 = c2, c1

    progress = tqdm(desc="Matching cells", total=len(c1), unit="cells")
    __progress_update.updater = progress.update

    # for each index corresponding to c1, returns the index in c2 that matches
    try:
        assignment = match_points(c1, c2, threshold, pre_match)
    finally:
        __progress_update.updater = None

    if progress is not None:
        progress.close()

    missing_c1, good_matches, missing_c2 = analyze_point_matches(
        c1, c2, assignment, threshold
    )
    if flip:
        missing_c1, missing_c2 = missing_c2, missing_c1
        good_matches = np.flip(good_matches, axis=1)
        good_matches = good_matches[good_matches[:, 0].argsort()]

    return missing_c1.tolist(), good_matches.tolist(), missing_c2.tolist()


# terrible hack. But you can't pass arbitrary objects to a njit function. But,
# it can access global variables and run them in objmode. So pass the progress
# updater to match_points via this global variable and function. We make it
# thread safe nominally, but it's not safe to modify within a thread while
# match_points is running

__progress_update = threading.local()
__progress_update.updater = None


def __compare_progress(n: int = 1) -> None:
    """Updates the progress bar by `n`, if there's one set."""
    if __progress_update.updater is not None:
        __progress_update.updater(n)


@njit
def _find_pairs_sorted(
    pos1: np.ndarray, pos2: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Given two sorted arrays, returns all the pairs in the arrays (without
    replacement) that are at a `np.isclose` distance of each other.

    This is computed in O(N) time.

    Parameters
    ----------
    pos1 : Sorted (1st axis) np.ndarray of shape `MxK`.
    pos2 : Sorted (1st axis) np.ndarray of shape `NxK`.

    Returns
    -------
    tuple :
        used_mask_1 : Bool np.ndarray of size `M`.
            It's True at indices of the rows in `pos1` used up in the pairing.
        used_mask_2 : Bool np.ndarray of size `N`.
            It's True at indices of the rows in `pos2` used up in the pairing.
        paired_indices: np.ndarray of shape `Rx2`.
            Each row is a pair of indices to `pos1` and `pos2`, respectively.
            Indicating a pair that is `close` to each other.
    """
    # mask of pos1/pos2 for the elements used in a identical par
    used_mask_1 = np.zeros(pos1.shape[0], dtype=np.bool_)
    used_mask_2 = np.zeros(pos2.shape[0], dtype=np.bool_)
    n_cols = pos1.shape[1]

    # the pos1/pos2 indices for each pair - at most this many pairs
    max_n = min(pos1.shape[0], pos2.shape[0])
    paired_indices = np.zeros((max_n, 2), dtype=np.int64)

    # how many pairs found
    used_n = 0
    # next index to check for pair for pos1/2
    pos1_i = 0
    pos2_i = 0

    # do this in O(N), until we reach end of either array
    while pos1_i < max_n and pos2_i < max_n:
        # are the two points the same
        same = True
        for i in range(n_cols):
            same = same and np.isclose(pos1[pos1_i, i], pos2[pos2_i, i])

        # they match
        if same:
            used_mask_1[pos1_i] = True
            used_mask_2[pos2_i] = True
            paired_indices[used_n, 0] = pos1_i
            paired_indices[used_n, 1] = pos2_i
            used_n += 1
            pos1_i += 1
            pos2_i += 1
        else:
            # the points are not the same in at least one dim, which is less?
            one_is_less = True
            for i in range(n_cols):
                # for dims until this one (if any), they are the same
                if pos1[pos1_i, i] < pos2[pos2_i, i]:  # first is less
                    break
                elif pos1[pos1_i, i] > pos2[pos2_i, i]:  # second is less
                    one_is_less = False
                    break
                # they were the same in this axis as well
            else:  # pragma: no cover
                assert False, "at least in one dim it should be different"

            if one_is_less:
                # first is less than second, advance first by one
                pos1_i += 1
            else:
                # second is less than first, advance second by one
                pos2_i += 1

    return used_mask_1, used_mask_2, paired_indices[:used_n, :]


def _find_identical_points(
    pos1: np.ndarray, pos2: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Given two arrays, returns the set of pairs of points in the arrays
    (without replacement) that are at a `np.isclose` distance of each other.

    This is computed in O(NlogN) time, dominated by sorting (internally).

    Parameters
    ----------
    pos1 : np.ndarray of shape `MxK`.
    pos2 : np.ndarray of shape `NxK`.

    Returns
    -------
    tuple :
        unpaired1_indices : np.ndarray of size `M - R`.
            Array of indices of `pos1` that were not used in any pairs
        unpaired2_indices : np.ndarray of size `N - R`.
            Array of indices of `pos2` that were not used in any pairs.
        paired_indices: np.ndarray of shape `Rx2`.
            Each row is a pair of indices to `pos1` and `pos2`, respectively.
            Indicating a pair that is `close` to each other.
    """
    # sort pos1 and pos2 rows, ordered by columns 1..N.
    # lexsort uses rows as keys and sorts the keys from last to first.
    # So flip order of rows and then transpose
    indices = np.lexsort(np.flip(pos1, axis=1).transpose())
    pos1_sorted = pos1[indices, :]
    # original pos1 indices of the sorted elements
    orig1_indices = np.arange(len(pos1), dtype=np.int64)[indices]

    indices = np.lexsort(np.flip(pos2, axis=1).transpose())
    pos2_sorted = pos2[indices, :]
    orig2_indices = np.arange(len(pos2), dtype=np.int64)[indices]

    # get the zero distance pairs
    used_mask_1, used_mask_2, paired_indices = _find_pairs_sorted(
        pos1_sorted, pos2_sorted
    )

    # convert the indices back to the original unsorted indices
    unpaired1_indices = orig1_indices[np.logical_not(used_mask_1)]
    unpaired2_indices = orig2_indices[np.logical_not(used_mask_2)]
    paired_indices[:, 0] = orig1_indices[paired_indices[:, 0]]
    paired_indices[:, 1] = orig2_indices[paired_indices[:, 1]]

    return unpaired1_indices, unpaired2_indices, paired_indices


@njit
def _optimize_pairs(
    pos1: np.ndarray,
    pos2: np.ndarray,
    threshold: float,
) -> np.ndarray:
    """
    Implements `match_points` using
    https://en.wikipedia.org/wiki/Hungarian_algorithm.

    Parameters
    ----------
    pos1 : np.ndarray
        2D array of NxK.
    pos2 : np.ndarray
        2D array of MxK.

        The relationship N <= M must be true and K must be the same for both.
    threshold : float, optional. Defaults to np.inf.
        The threshold to use to consider a pair a bad match. Any match pair
        whose distance is greater or equal to the threshold will be considered
        to be at threshold distance (i.e. max distance).

        It'll still show up in the matching, but it will have the least
        priority for a match because that match will not reduce the overall
        cost across all points.

    Returns
    -------
    matches : np.ndarray
        1D array of length N, where index i in matches corresponds
        to index i in `pos1` and its value is the index in pos2
        of its best match.
    """
    # we don't check for boundary conditions, just assert because it should be
    # checked by caller (match_points)
    n_rows = pos1.shape[0]
    n_cols = pos2.shape[0]
    assert len(pos1.shape) == 2
    assert len(pos2.shape) == 2
    assert pos1.shape[1] == pos2.shape[1]
    assert n_rows <= n_cols

    pos1 = pos1.astype(np.float64)
    pos2 = pos2.astype(np.float64)

    have_threshold = threshold != np.inf

    potentials_rows = np.zeros(n_rows)
    potentials_cols = np.zeros(n_cols + 1)
    assignment_row = np.full(n_cols + 1, -1, dtype=np.int_)
    min_to = np.empty(n_cols + 1, dtype=np.float64)
    # previous worker on alternating path
    prev_col_for_col = np.empty(n_cols + 1, dtype=np.int_)
    # whether col is in use
    col_used = np.zeros(n_cols + 1, dtype=np.bool_)

    # assign row-th match
    for row in range(n_rows):
        col = n_cols
        assignment_row[col] = row
        # min reduced cost over edges from Z to worker w
        min_to[:] = np.inf
        prev_col_for_col[:] = -1
        col_used[:] = False

        # runs at most row + 1 times
        while assignment_row[col] != -1:
            col_used[col] = True
            row_cur = assignment_row[col]
            delta = np.inf
            col_next = -1

            for col_i in range(n_cols):
                if not col_used[col_i]:
                    # use sqrt to match threshold which is in actual distance
                    dist = 0.0
                    for i in range(pos1.shape[1]):
                        dist += math.pow(pos1[row_cur, i] - pos2[col_i, i], 2)
                    dist = math.sqrt(dist)
                    if dist == np.inf:
                        raise ValueError(
                            "The distance between points is too large"
                        )
                    if have_threshold and dist > threshold:
                        dist = threshold

                    cur = (
                        dist
                        - potentials_rows[row_cur]
                        - potentials_cols[col_i]
                    )
                    if cur < min_to[col_i]:
                        min_to[col_i] = cur
                        prev_col_for_col[col_i] = col

                    if min_to[col_i] < delta:
                        delta = min_to[col_i]
                        col_next = col_i

            # delta will always be non-negative,
            # except possibly during the first time this loop runs
            # if any entries of C[row] are negative
            for col_i in range(n_cols + 1):
                if col_used[col_i]:
                    potentials_rows[assignment_row[col_i]] += delta
                    potentials_cols[col_i] -= delta
                else:
                    min_to[col_i] -= delta
            col = col_next

        # update assignments along alternating path
        while col != n_cols:
            col_i = prev_col_for_col[col]
            assignment_row[col] = assignment_row[col_i]
            col = col_i

        with objmode():
            __compare_progress()

    # compute match from assignment
    matches = np.empty(n_rows, dtype=np.int64)
    for i in range(n_cols):
        if assignment_row[i] != -1:
            matches[assignment_row[i]] = i

    return matches


def match_points(
    pos1: np.ndarray,
    pos2: np.ndarray,
    threshold: float = np.inf,
    pre_match: bool = True,
) -> np.ndarray:
    """
    Given two arrays, each a list of position. For each point in `pos1` it
    finds a point in `pos2` such that the distance between the assigned
    matches across all `pos1` is minimized.

    E.g.::

        >>> pos1 = np.array([[20, 10, 30, 40]]).T
        >>> pos2 = np.array([[5, 15, 25, 35, 50]]).T
        >>> matches = match_points(pos1, pos2)
        >>> matches
        array([1, 0, 2, 3])

    Parameters
    ----------
    pos1 : np.ndarray
        2D array of NxK. Where N is number of positions and K is the number
        of dimensions (e.g. 3 for x, y, z).
    pos2 : np.ndarray
        2D array of MxK. Where M is number of positions and K is the number
        of dimensions (e.g. 3 for x, y, z).

        The relationship N <= M must be true.
    threshold : float, optional. Defaults to np.inf.
        The threshold to use to consider a pair a bad match. Any match pair
        whose distance is greater or equal to the threshold will be considered
        to be at threshold distance (i.e. the max distance).

        It'll still show up in the matching, but it will have the least
        priority for a match because that match will not reduce the overall
        cost across all points.

        Use `analyze_point_matches` with the same threshold subsequently to
        remove the "bad" matches.
    pre_match : bool, optional. Defaults to True.
        If True, we will (interenally) first efficiently find all the pairs of
        `pos1` and `pos2` which are each at the same position in space. Then
        we run the optimization to find the best matching on the remaining.

        If True, it'll significantly speed up the matching, if there are pairs
        of points on top of each other across the input lists.

    Returns
    -------
    matches : np.ndarray
        1D array of length N. Each index i in matches corresponds
        to index i in `pos1`. The value of index i in matches is the index
        j in pos2 that is the best match for that pos1.

        I.e. the match is (pos1[i], pos2[matches[i]]).
    """
    if len(pos1.shape) != 2 or len(pos2.shape) != 2:
        raise ValueError("The input arrays must have exactly 2 dimensions")

    n_rows = pos1.shape[0]
    n_cols = pos2.shape[0]
    if n_rows > n_cols:
        raise ValueError(
            "The length of pos1 must be less than or equal to length of pos2"
        )
    if pos1.shape[1] != pos2.shape[1]:
        raise ValueError("The two inputs have different number of columns")

    if not pre_match:
        # do optimization on full inputs
        return _optimize_pairs(pos1, pos2, threshold)

    # extract the indices of zero-pairs and remaining points
    unpaired1_indices, unpaired2_indices, paired_indices = (
        _find_identical_points(pos1, pos2)
    )
    # the number of zero pairs found are done!
    __compare_progress(len(paired_indices))

    # if everything was zero pairs, we're done!
    if not len(unpaired1_indices):
        # sort by pos1 and then return the corresponding pos2 indices matching
        # those pos1 points
        pos1_sorted_indices = np.argsort(paired_indices[:, 0])
        return paired_indices[pos1_sorted_indices, 1]

    # extract remaining pos1/post2 and run optimization
    pos1 = pos1[unpaired1_indices]
    pos2 = pos2[unpaired2_indices]
    n_rows = pos1.shape[0]

    matches = _optimize_pairs(pos1, pos2, threshold)

    # map extracted
    full_matches = np.empty(n_rows + len(paired_indices), dtype=np.int64)
    # set pos1 optimized matches to their pos2 indices
    full_matches[unpaired1_indices] = unpaired2_indices[matches]
    # set the zero pairs pos1 to corresponding pos2 match indices
    full_matches[paired_indices[:, 0]] = paired_indices[:, 1]

    return full_matches


@njit
def analyze_point_matches(
    pos1: np.ndarray,
    pos2: np.ndarray,
    matches: np.ndarray,
    threshold: float = np.inf,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Given a matching found by `match_points`, it optionally applies a threshold
    and splits the matched points from unmatched points in a friendlier way.

    E.g.::

        >>> pos1 = np.array([[20, 10, 30, 40, 50]]).T
        >>> pos2 = np.array([[5, 15, 25, 35, 100, 200]]).T
        >>> matches = match_points(pos1, pos2)
        >>> matches
        array([1, 0, 2, 3, 4])
        >>> analyze_point_matches(pos1, pos2, matches)
        (array([], dtype=int64),
         array([[0, 1],
                [1, 0],
                [2, 2],
                [3, 3],
                [4, 4]], dtype=int64),
         array([5], dtype=int64))
         >>> analyze_point_matches(pos1, pos2, matches, threshold=10)
        (array([4], dtype=int64),
         array([[0, 1],
                [1, 0],
                [2, 2],
                [3, 3]], dtype=int64),
         array([4, 5], dtype=int64))

    Parameters
    ----------
    pos1 : np.ndarray
        Same as `match_points`.
    pos2 : np.ndarray
        Same as `match_points`.
    matches : np.ndarray
        The matches returned by `match_points`.
    threshold : float, optional. Defaults to np.inf.
        The threshold to use to remove bad matches. Any match pair whose
        distance is greater than the threshold will be removed from the
        matching and added to the missing_pos1 and missing_pos2 arrays.

        To get a best global optimum, use the same threshold you used in
        `match_points`.

    Returns
    -------
    tuple : (np.ndarray, np.ndarray, np.ndarray)
        missing_pos1: 1d array of all the indices of pos1 that found no match
            in pos2 (sorted).
        good_matches: 2d array with all the (pos1, pos2) indices that remained
            in the matching. It's of size Rx2. It's sorted by the first column.
        missing_pos2: 1d array of all the indices of pos2 that found no match
            in pos1 (sorted).
    """
    # indices and mask on indices
    pos2_n = len(pos2)
    pos2_i = np.arange(pos2_n)
    pos2_unmatched = np.ones(pos2_n, dtype=np.bool_)
    # those in pos2 who have a match in pos1
    pos2_unmatched[matches] = False
    # all the pos2 that have no matches from pos1
    missing_pos2 = pos2_i[pos2_unmatched]

    # repackage matches so the first column is the pos1 idx and 2nd column is
    # the corresponding pos2 index
    matches_indices = np.stack((np.arange(len(pos1)), matches), axis=1)

    dist = np.sqrt(np.sum(np.square(pos1 - pos2[matches, :]), axis=1))
    too_large = dist >= threshold
    bad_matches = matches_indices[too_large, :]
    good_matches = matches_indices[np.logical_not(too_large), :]

    missing_pos1 = bad_matches[:, 0]
    # more missing for pos2 for those above threshold
    missing_pos2 = np.concatenate((missing_pos2, bad_matches[:, 1]))
    missing_pos2 = np.sort(missing_pos2)

    return missing_pos1, good_matches, missing_pos2
